@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -24,6 +24,12 @@
 
 #define CHANNEL_STORE_OPCODE_LIMIT        256
 
+//
+// One split store for each resource.
+//
+
+#define CHANNEL_STORE_RESOURCE_LIMIT      256
+
 class ChannelStore
 {
   public:
