@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -23,36 +23,49 @@
 #include "Misc.h"
 #include "Types.h"
 #include "Timestamp.h"
+#include "Statistics.h"
 
 #define PANIC
 #define WARNING
 #undef  TEST
 #undef  DEBUG
 
-//
-// Sensible default values set at compile time.
-// User can override them by providing a control
-// file in his/her .nx directory.
-//
-
-extern const char ControlModem[];
-extern const char ControlIsdn[];
-extern const char ControlAdsl[];
-extern const char ControlLocal[];
-
 //
 // This is the mode proxy is running.
 //
 
 typedef enum
 {
-  PROXY_UNDEFINED = -1,
-  PROXY_CLIENT,
-  PROXY_SERVER,
-  PROXY_LAST_TAG
+  proxy_undefined = -1,
+  proxy_client,
+  proxy_server,
+  proxy_last_tag
 }
 T_proxy_mode;
 
+//
+// Handle advances in the connection
+// procedure.
+//
+
+typedef enum
+{
+  stage_undefined,
+  stage_initializing,
+  stage_connecting,
+  stage_connected,
+  stage_waiting_forwarder_version,
+  stage_waiting_forwarder_options,
+  stage_sending_forwarder_options,
+  stage_waiting_proxy_version,
+  stage_waiting_proxy_options,
+  stage_sending_proxy_options,
+  stage_waiting_proxy_caches,
+  stage_sending_proxy_caches,
+  stage_operational
+}
+T_proxy_stage;
+
 //
 // This is a suggestion about the type
 // of session that's running.
@@ -60,21 +73,34 @@ T_proxy_mode;
 
 typedef enum
 {
-  SESSION_UNDEFINED = -1,
-  SESSION_X,
-  SESSION_RDP,
-  SESSION_RFB,
-  SESSION_APPLICATION,
-  SESSION_LAST_TAG
+  session_undefined = -1,
+  session_x,
+  session_rdp,
+  session_rfb,
+  session_raw,
+  session_last_tag
 }
 T_session_mode;
 
 //
-// Link mode will be set to values from
-// NX proto.
+// Set how data will be written to the peer
+// socket.
 //
 
-#define LINK_UNDEFINED             -1;
+typedef enum
+{
+  policy_undefined = -1,
+  policy_immediate,
+  policy_deferred
+}
+T_flush_policy;
+
+//
+// Link mode, after negotiation, will be set to
+// any of the values defined in the NXproto.h.
+//
+
+#define link_undefined  -1;
 
 //
 // This class collects functioning parameters, 
@@ -96,12 +122,35 @@ class Control
 
   T_proxy_mode ProxyMode;
 
+  //
+  // Goes from initializing to operational.
+  //
+
+  T_proxy_stage ProxyStage;
+
   //
   // Hint about type of session currently running.
   //
 
   T_session_mode SessionMode;
 
+  //
+  // Either immediate or defferred flushes.
+  //
+
+  T_flush_policy FlushPolicy;
+
+  //
+  // If set, the channels will try to flush the
+  // encoded data whenever there is a prioritized
+  // message. Depending on the flush policy, this
+  // may determine an immediate flush or an event
+  // being generated telling to the agent that it
+  // should flush the proxy link.
+  //
+
+  int FlushPriority;
+
   //
   // Id corresponding to link speed negotiated
   // between proxies.
@@ -109,6 +158,43 @@ class Control
 
   int LinkMode;
 
+  //
+  // Set if the proxy is connected to a program
+  // providing the encryption of the point to
+  // point communication.
+  //
+
+  int LinkEncrypted;
+
+  //
+  // Maximum number of bytes sent for each token.
+  //
+
+  int TokenSize;
+
+  //
+  // Maximum number of tokens that can be spent
+  // by the client proxy before having to block
+  // waiting for a reply.
+  //
+
+  int TokenLimit;
+
+  //
+  // Report a congestion if the tokens come below
+  // this threshold.
+  //
+
+  int TokenThreshold;
+
+  //
+  // Bitmask used to determine the distribution
+  // of channel ids between the client and server
+  // proxies.
+  //
+
+  int ChannelMask;
+
   //
   // Kill session if control parameters cannot
   // be negotiated before this timeout.
@@ -116,13 +202,6 @@ class Control
 
   int InitTimeout;
 
-  //
-  // Timeout to be used to slow down the most
-  // demanding agents' clients.
-  //
-
-  int WakeupTimeout;
-
   //
   // How many milliseconds to wait before to restart
   // handling of pending messages after a timeout.
@@ -130,13 +209,6 @@ class Control
 
   int PendingTimeout;
 
-  //
-  // Schedule writes to remote proxy by enqueuing
-  // data for this amount of milliseconds.
-  //
-
-  int FlushTimeout;
-
   //
   // Enqueue motion notify events in server channel.
   //
@@ -144,20 +216,21 @@ class Control
   int MotionTimeout;
 
   //
-  // Send a ping to remote proxy if no other data
-  // was sent for this amount of milliseconds.
+  // Agents can try to keep the frequency of packets
+  // constant by sending a null request if no other
+  // data was produced in the given timeframe.
+  //
+
+  int FrameTimeout;
+
+  //
+  // Enter the congestion state if the remote does
+  // not reply to the ping within the given amount
+  // of time.
   //
 
   int PingTimeout;
 
-  //
-  // How much time to wait before reading again
-  // from proxy when any channel has surpassed
-  // its buffer limit.
-  //
-
-  int LimitTimeout;
-
   //
   // Timeout used by server channel at reset.
   //
@@ -165,11 +238,11 @@ class Control
   int ResetTimeout;
 
   //
-  // Close connection if can't write before
+  // Close the connection if can't write before
   // this timeout.
   //
 
-  int WriteTimeout;
+  int ChannelTimeout;
 
   //
   // Close connection if can't write before
@@ -179,8 +252,8 @@ class Control
   int ProxyTimeout;
 
   //
-  // Minimum timeout to be used to check state of
-  // proxy link.
+  // Timeout used to wait for a complete message
+  // to become available.
   //
 
   int RetryTimeout;
@@ -209,13 +282,6 @@ class Control
 
   int CleanupTimeout;
 
-  //
-  // Slow down the client proxy if X server can't
-  // keep up handling X messages.
-  //
-
-  int CongestionTimeout;
-
   //
   // Wait this amount of milliseconds before any
   // iteration of the house-keeping process.
@@ -230,19 +296,28 @@ class Control
   int LatencyTimeout;
 
   //
-  // Maximum allowed size of log file.
+  // Maximum allowed size of log files.
   //
 
-  int LogFileSizeLimit;
+  int FileSizeLimit;
   int FileSizeCheckTimeout;
 
   //
-  // Do we quit proxy in case of failure or
-  // we let it drop the old connection and
-  // restart?
+  // What do we do at the end of session? If
+  // this flag is set we launch a new client
+  // letting the user run a new NX session.
   //
 
-  int EnableRestartOnFailure;
+  int EnableRestartOnShutdown;
+
+  //
+  // The client can request the proxy to kill
+  // a number of processes before exiting.
+  //
+
+  int *KillDaemonOnShutdown;
+  int KillDaemonOnShutdownNumber;
+  int KillDaemonOnShutdownLimit;
 
   //
   // Do we generate a core dump and exit in
@@ -252,16 +327,10 @@ class Control
   int EnableCoreDumpOnAbort;
 
   //
-  // What to do in case of SIGHUP. By default
-  // we close down the proxy link and perform
-  // a clean shutdown of the session. You can
-  // even re-read the session configuration
-  // or restart the proxy link.
+  // Is statistic output enabled?
   //
 
-  int EnableShutdownOnSighup;
-  int EnableReconfigOnSighup;
-  int EnableRestartOnSighup;
+  int EnableStatistics;
 
   //
   // Version number of local and remote proxy.
@@ -275,6 +344,10 @@ class Control
   int RemoteVersionMinor;
   int RemoteVersionPatch;
 
+  int CompatVersionMajor;
+  int CompatVersionMinor;
+  int CompatVersionPatch;
+
   //
   // Which unpack methods are implemented in proxy?
   //
@@ -350,9 +423,9 @@ class Control
   int TransportProxyBufferSize;
   int TransportGenericBufferSize;
 
-  int TransportXBufferThreshold;
-  int TransportProxyBufferThreshold;
-  int TransportGenericBufferThreshold;
+  int TransportXBufferLimit;
+  int TransportProxyBufferLimit;
+  int TransportGenericBufferLimit;
 
   int TransportMaximumBufferSize;
   int TransportForceBufferSize;
@@ -362,9 +435,9 @@ class Control
   // exceed these buffer limits.
   //
 
-  int TransportXBufferLimit;
-  int TransportProxyBufferLimit;
-  int TransportGenericBufferLimit;
+  int TransportXBufferThreshold;
+  int TransportProxyBufferThreshold;
+  int TransportGenericBufferThreshold;
 
   //
   // Immediately flush data produced for the
@@ -373,13 +446,6 @@ class Control
 
   int TransportWriteThreshold;
 
-  //
-  // Immediately flush if shared memory data
-  // exceeds this limit.
-  // 
-
-  int TransportShmemThreshold;
-
   //
   // Set this to the maximum amount of data that 
   // will be sent in a single write operation.
@@ -433,39 +499,16 @@ class Control
   int ClientProcessorLimit;
   int ServerProcessorLimit;
   
-  //
-  // How many bytes to write and how many
-  // milliseconds to consume handling mes-
-  // sages before yielding control to main
-  // loop.
-  //
-
-  int LocalTimeInARowLimit;
-  int LocalBytesInARowLimit;
-  int LocalOutputInARowLimit;
-
-  int ClientTimeInARowLimit;
-  int ClientBytesInARowLimit;
-  int ClientReadsInARowLimit;
-  int ClientMessagesInARowLimit;
-  int ClientOutputInARowLimit;
-
-  int ServerTimeInARowLimit;
-  int ServerBytesInARowLimit;
-  int ServerReadsInARowLimit;
-  int ServerMessagesInARowLimit;
-  int ServerOutputInARowLimit;
-
   //
   // This is the limit imposed by user on
   // total cache size.
   //
 
-  int ClientTotalStorageSizeLimit;
-  int ServerTotalStorageSizeLimit;
+  int ClientTotalStorageSize;
+  int ServerTotalStorageSize;
 
-  int LocalTotalStorageSizeLimit;
-  int RemoteTotalStorageSizeLimit;
+  int LocalTotalStorageSize;
+  int RemoteTotalStorageSize;
 
   //
   // Discard messages in store older than
@@ -512,108 +555,106 @@ class Control
   // Directives on size of messages to cache.
   //
 
-  int MinimumMessageSizeThreshold;
-  int MaximumMessageSizeThreshold;
+  int MinimumMessageSize;
+  int MaximumMessageSize;
 
   //
-  // Any given amount of milliseconds send
-  // a new split to the remote side.
+  // Maximum size of a single X request.
+  //
+
+  int MaximumRequestSize;
+
+  //
+  // Currently selected streaming mode.
+  //
+
+  int SplitMode;
+
+  //
+  // Send new split data any given amount of
+  // milliseconds.
   //
 
   int SplitTimeout;
 
   //
-  // Split messages are sent asynchronously
-  // on the link. Messages smaller that this
-  // threshold are sent in one shot, and the
-  // agent is immediately notified.
+  // Maximum number of distinct messages and
+  // maximum size in bytes of the temporary
+  // storage.
+  //
+
+  int SplitTotalSize;
+  int SplitTotalStorageSize;
+
+  //
+  // Don't split messages smaller that this
+  // threshold and send no more than the
+  // given amount of bytes in a single data
+  // shot when streaming the split messages.
   //
 
   int SplitDataThreshold;
-
-  //
-  // This is maximum block size to be used to
-  // transfer split messages to the remote
-  // peer.
-  //
-
   int SplitDataPacketLimit;
 
   //
-  // Maximum size of temporary storage.
+  // Agent related parameters. These values apply
+  // to the agent which, at startup, must query
+  // the user's settings.
   //
 
-  int SplitTotalStorageSizeLimit;
-
-  //
-  // Is statistic output enabled?
-  //
-
-  int CollectStatistics;
-
-  //
-  // Agent related parameters. Some of these
-  // values apply to agent which, at startup,
-  // must query the user's settings, others
-  // are related to internal optimizations
-  // when running proxy in an NX environment.
-  //
-
-  int AgentKarmaSize;
-  int AgentSplitSize;
-  int AgentPackMethod;
-  int AgentPackQuality;
-  int AgentFlushPriority;
-  int AgentFlushImmediate;
-  int AgentSyncPropagate;
-  int AgentForceReliable;
-  int AgentHideRender;
-  int AgentTaintReplies;
-  int AgentTaintThreshold;
+  int PackMethod;
+  int PackQuality;
+  int HideRender;
+  int TaintReplies;
+  int TaintThreshold;
 
   //
   // Do we allow shared memory image support in
   // client and or server?
   //
 
-  int AgentShmemClient;
-  int AgentShmemServer;
+  int ShmemClient;
+  int ShmemServer;
 
   //
   // Default size of shared memory segments used
   // in MIT-SHM support.
   //
 
-  int AgentShmemClientSize;
-  int AgentShmemServerSize;
+  int ShmemClientSize;
+  int ShmemServerSize;
 
   //
-  // Should agent apply clean of pad and unused
-  // bytes in X protocol?
+  // The user's home directory.
   //
 
-  int AgentCleanGet;
-  int AgentCleanAlloc;
-  int AgentCleanFlush;
-  int AgentCleanSend;
-  int AgentCleanImages;
+  char *HomePath;
 
   //
-  // Should apply special operation on images?
-  //
-
-  int AgentImageSplit;
-  int AgentImageMask;
-  int AgentImageFrame;
-  int AgentImageSplitMethod;
-  int AgentImageMaskMethod;
-
-  //
-  // Usually the user's ".nx" directory.
+  // The ".nx" directory, usually in
+  // the user's home.
   //
 
   char *RootPath;
 
+  //
+  // Usually the /usr/NX" directory.
+  //
+
+  char *SystemPath;
+
+  //
+  // Usually the "/tmp" directory.
+  //
+
+  char *TempPath;
+
+  //
+  // The complete path to the client.
+  //
+
+  char *ClientPath;
+
   //
   // String containing path of cache
   // file selected for load or save.
@@ -641,6 +682,14 @@ class Control
   int PersistentCacheEnableLoad;
   int PersistentCacheEnableSave;
 
+  //
+  // This is used just for test because
+  // it requires that client and server
+  // reside on the same machine.
+  //
+
+  int PersistentCacheCheckOnShutdown;
+
   //
   // Load packed image and render extension
   // message stores. This currently depends
@@ -651,8 +700,8 @@ class Control
   int PersistentCacheLoadRender;
 
   //
-  // Maximum allowed size of image
-  // cache on disk.
+  // Maximum disk consumption of message
+  // caches on disk.
   //
 
   int PersistentCacheDiskLimit;
@@ -672,8 +721,8 @@ class Control
   int ImageCacheEnableSave;
 
   //
-  // Maximum allowed size of image
-  // cache on disk.
+  // Maximum disk consumption of image
+  // caches on disk.
   //
 
   int ImageCacheDiskLimit;
@@ -693,20 +742,20 @@ class Control
   // control, this info must be moved elsewhere.
   //
 
-  int getUpperStorageSizeLimit() const
+  int getUpperStorageSize() const
   {
-    return (ClientTotalStorageSizeLimit >
-                ServerTotalStorageSizeLimit ?
-                    ClientTotalStorageSizeLimit :
-                        ServerTotalStorageSizeLimit);
+    return (ClientTotalStorageSize >
+                ServerTotalStorageSize ?
+                    ClientTotalStorageSize :
+                        ServerTotalStorageSize);
   }
 
-  int getLowerStorageSizeLimit() const
+  int getLowerStorageSize() const
   {
-    return (ClientTotalStorageSizeLimit <
-                ServerTotalStorageSizeLimit ?
-                    ClientTotalStorageSizeLimit :
-                        ServerTotalStorageSizeLimit);
+    return (ClientTotalStorageSize <
+                ServerTotalStorageSize ?
+                    ClientTotalStorageSize :
+                        ServerTotalStorageSize);
   }
 
   //
@@ -783,170 +832,15 @@ class Control
   }
 
   //
-  // Track per-frame time and bitrate
-  // usage.
+  // Update the current bitrate.
   //
 
-  void resetTimeInARow()
-  {
-    timeInARow_ = getTimestamp();
-  }
-
-  int getTimeInARow()
-  {
-    return diffTimestamp(timeInARow_, getTimestamp());
-  }
-
-  void addBitsInARow(int numBits)
-  {
-    bitsInARow_ += numBits;
-  }
-
-  int getBytesInARow()
-  {
-    return (bitsInARow_ >> 3);
-  }
-
-  void resetBytesInARow(int initialBytes = 0)
-  {
-    bitsInARow_ = (initialBytes << 3);
-  }
+  void updateBitrate(int bytes);
 
   //
-  // Track how many bytes are written
-  // to X, SMB, Audio, Sync and other
-  // servers running at decoding side.
+  // Return the current bitrate.
   //
 
-  void addOutputInARow(int numBytes)
-  {
-    outputInARow_ += numBytes;
-  }
-
-  void removeOutputInARow(int numBytes)
-  {
-    outputInARow_ -= numBytes;
-  }
-
-  int getOutputInARow()
-  {
-    return outputInARow_;
-  }
-
-  void resetOutputInARow()
-  {
-    outputInARow_ = 0;
-  }
-
-  //
-  // Track bytes to be written in each
-  // chunck of split data.
-  //
-
-  int getBytesToSplitInARow()
-  {
-    //
-    // It should take in account the
-    // current bitrate and adapt bytes
-    // to send according to it.
-    //
-
-    return SplitDataPacketLimit;
-  }
-
-  int getTimeToNextSplit()
-  {
-    long timeToSplit = SplitTimeout -
-                          diffTimestamp(lastSplitTs_, getTimestamp());
-
-    #ifdef TEST
-    *logofs << "Control: Time to next split is " << timeToSplit
-            << " with timeout of " << SplitTimeout << " Ms.\n"
-            << logofs_flush;
-    #endif
-
-    if (timeToSplit > 0)
-    {
-      return timeToSplit;
-    }
-
-    return 0;
-  }
-
-  int isTimeToSplit()
-  {
-    return ((SplitTimeout - LatencyTimeout -
-                diffTimestamp(lastSplitTs_, getTimestamp())) <= 0);
-  }
-
-  void setLastSplitInfo()
-  {
-    lastSplitTs_ = getTimestamp();
-  }
-
-  int getTimeToNextFlush()
-  {
-    long timeToFlush = FlushTimeout -
-                          diffTimestamp(lastFlushTs_, getTimestamp());
-
-    #ifdef TEST
-    *logofs << "Control: Time to next flush is " << timeToFlush
-            << " with timeout of " << FlushTimeout << " Ms.\n"
-            << logofs_flush;
-    #endif
-
-    if (timeToFlush > 0)
-    {
-      return timeToFlush;
-    }
-
-    return 0;
-  }
-
-  int getBytesToNextFlush()
-  {
-    int bytesToFlush = LocalBytesInARowLimit - getBytesInARow();
-
-    if (bytesToFlush < 0)
-    {
-      bytesToFlush = 0;
-    }
-
-    return bytesToFlush;
-  }
-
-  //
-  // This rounds up to the value of latency
-  // timeout so you could get a valid time
-  // from the previous function even if it
-  // is already time to flush.
-  // 
-
-  int isTimeToFlush()
-  {
-    return ((FlushTimeout - LatencyTimeout -
-                diffTimestamp(lastFlushTs_, getTimestamp())) <= 0);
-  }
-
-  void setLastFlushInfo()
-  {
-    lastFlushTs_ = getTimestamp();
-  }
-
-  //
-  // Check if handling of messages took more
-  // time than advisable or data produced for
-  // peer proxy is enough make a full write.
-  //
-
-  int isTimeToYield(T_yield type);
-
-  //
-  // Calculate current bitrate.
-  //
-
-  void setLastWriteInfo(int bytes);
-
   int getBitrateInShortFrame()
   {
     return bitrateInShortFrame_;
@@ -967,6 +861,11 @@ class Control
     topBitrate_ = 0;
   }
 
+  double getStreamRatio()
+  {
+    return statistics -> getStreamRatio();
+  }
+
   int isStartup()
   {
     if (isStartup_ == 0)
@@ -978,7 +877,8 @@ class Control
       if (StartupTimeout >= LatencyTimeout)
       {
         isStartup_ = ((StartupTimeout - LatencyTimeout -
-                           diffTimestamp(startupTs_, getTimestamp())) > 0);
+                          diffTimestamp(startupTs_,
+                              getTimestamp())) > 0);
       }
       else
       {
@@ -989,56 +889,36 @@ class Control
     return isStartup_;
   }
 
-  void setProtoStep1()
+  void setProtoStep6()
   {
-    protoStep1_ = 1;
-    protoStep2_ = 0;
-    protoStep3_ = 0;
-    protoStep4_ = 0;
+    protoStep6_ = 1;
+    protoStep7_ = 0;
   }
 
-  void setProtoStep2()
+  void setProtoStep7()
   {
-    protoStep1_ = 0;
-    protoStep2_ = 1;
-    protoStep3_ = 0;
-    protoStep4_ = 0;
+    protoStep6_ = 1;
+    protoStep7_ = 1;
   }
 
-  void setProtoStep3()
+  int isProtoStep6()
   {
-    protoStep1_ = 0;
-    protoStep2_ = 1;
-    protoStep3_ = 1;
-    protoStep4_ = 0;
+    return protoStep6_;
   }
 
-  void setProtoStep4()
+  int isProtoStep7()
   {
-    protoStep1_ = 0;
-    protoStep2_ = 1;
-    protoStep3_ = 1;
-    protoStep4_ = 1;
+    return protoStep7_;
   }
 
-  int isProtoStep1()
+  void setCacheStep3()
   {
-    return protoStep1_;
+    cacheStep3_ = 1;
   }
 
-  int isProtoStep2()
+  int isCacheStep3()
   {
-    return protoStep2_;
-  }
-
-  int isProtoStep3()
-  {
-    return protoStep3_;
-  }
-
-  int isProtoStep4()
-  {
-    return protoStep4_;
+    return cacheStep3_;
   }
 
   private:
@@ -1049,16 +929,6 @@ class Control
 
   void setLocalUnpackMethods();
 
-  //
-  // Track time spent handling messages
-  // and bits produced in the current
-  // run of the main loop.
-  //
-
-  T_timestamp timeInARow_;
-  int         bitsInARow_;
-  int         outputInARow_;
-
   //
   // Track CPU and network usage.
   //
@@ -1071,21 +941,13 @@ class Control
   double bytesOut_;
 
   //
-  // Event timestamp is actually a
-  // timeout, not a timestamp.
+  // Track the current bitrate used for
+  // I/O operations on the proxy link.
   //
 
-  T_timestamp lastSplitTs_;
-  T_timestamp lastFlushTs_;
-
-  //
-  // Track bitrate on proxy link.
-  //
-
-  T_timestamp lastWriteTs_;
-
   T_timestamp startShortFrameTs_;
   T_timestamp startLongFrameTs_;
+  T_timestamp startFrameTs_;
 
   int bytesInShortFrame_;
   int bytesInLongFrame_;
@@ -1094,7 +956,6 @@ class Control
   int bitrateInLongFrame_;
 
   int topBitrate_;
-  int upgradeBitrate_;
 
   //
   // Time since we started the proxy.
@@ -1108,10 +969,14 @@ class Control
   // protocol versions.
   //
 
-  int protoStep1_;
-  int protoStep2_;
-  int protoStep3_;
-  int protoStep4_;
+  int protoStep6_;
+  int protoStep7_;
+
+  //
+  // Manage the version of cache.
+  //
+
+  int cacheStep3_;
 };
 
 #endif /* Control_H */
