@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -18,6 +18,7 @@
 #include <X11/Xmd.h>
 
 #include <unistd.h>
+#include <setjmp.h>
 #include <zlib.h>
 
 #ifdef __cplusplus
@@ -45,17 +46,17 @@ extern "C"
 #undef  DEBUG
 
 #define RGB24_TO_PIXEL(bpp,r,g,b)                               \
-   ((((CARD##bpp)(r) & 0xFF) * srcRedMax + 127) / 255           \
+   ((((CARD##bpp)(r) & 0xff) * srcRedMax + 127) / 255           \
     << srcRedShift |                                            \
-    (((CARD##bpp)(g) & 0xFF) * srcGreenMax + 127) / 255         \
+    (((CARD##bpp)(g) & 0xff) * srcGreenMax + 127) / 255         \
     << srcGreenShift |                                          \
-    (((CARD##bpp)(b) & 0xFF) * srcBlueMax + 127) / 255          \
+    (((CARD##bpp)(b) & 0xff) * srcBlueMax + 127) / 255          \
     << srcBlueShift)
 
 #define RGB24_TO_PIXEL32(r,g,b)                                 \
-  (((CARD32)(r) & 0xFF) << srcRedShift |                        \
-   ((CARD32)(g) & 0xFF) << srcGreenShift |                      \
-   ((CARD32)(b) & 0xFF) << srcBlueShift)
+  (((CARD32)(r) & 0xff) << srcRedShift |                        \
+   ((CARD32)(g) & 0xff) << srcGreenShift |                      \
+   ((CARD32)(b) & 0xff) << srcBlueShift)
 
 //
 // Functions from Unpack.cpp
@@ -82,13 +83,13 @@ static void JpegSkipInputData(j_decompress_ptr cinfo, long num_bytes);
 static boolean JpegFillInputBuffer(j_decompress_ptr cinfo);
 
 static int DecompressJpeg16(unsigned char *compressedData, int compressedLen,
-                                unsigned int w, unsigned int h, unsigned short *dst_buf);
+                                unsigned int w, unsigned int h, unsigned char *dstBuf, int byteOrder);
 
 static int DecompressJpeg24(unsigned char *compressedData, int compressedLen,
-                                unsigned int w, unsigned int h, unsigned char *dst_buf);
+                                unsigned int w, unsigned int h, unsigned char *dstBuf, int byteOrder);
 
 static int DecompressJpeg32(unsigned char *compressedData, int compressedLen,
-                                unsigned int w, unsigned int h, unsigned int *dst_buf);
+                                unsigned int w, unsigned int h, unsigned char *dstBuf, int byteOrder);
 
 //
 // Colormap stuff.
@@ -103,16 +104,21 @@ CARD8  srcRedShift, srcGreenShift, srcBlueShift;
 
 static bool jpegError;
 
+jmp_buf UnpackJpegContext;
+
 void UnpackJpegErrorHandler(j_common_ptr cinfo)
 {
   #ifdef PANIC
-  *logofs << "UnpackJpeg: PANIC! Fatal error in JPEG decompression.\n"
+  *logofs << "UnpackJpegErrorHandler: PANIC! Detected error in JPEG decompression.\n"
+          << logofs_flush;
+
+  *logofs << "UnpackJpegErrorHandler: PANIC! Trying to revert to the previous context.\n"
           << logofs_flush;
   #endif
 
-  cerr << "Error" << ": Fatal error in JPEG decompression.\n";
+  jpegError = 1;
 
-  HandleAbort();
+  longjmp(UnpackJpegContext, 1);
 }
 
 //
@@ -127,17 +133,18 @@ static char    *tmpBuf;
 static int     tmpBufSize = 0;
 
 int UnpackJpeg(T_geometry *geometry, unsigned char method,
-                   unsigned char *src_data, int src_size, int dst_bpp,
-                       int dst_width, int dst_height, unsigned char *dst_data,
-                           int dst_size)
+                   unsigned char *srcData, int srcSize, int dstBpp,
+                       int dstWidth, int dstHeight, unsigned char *dstData,
+                           int dstSize, int byteOrder)
 {
   //
   // Check if data is coming from a failed unsplit.
   //
 
-  if (src_size < 2 || src_data[0] == 0x88 && src_data[1] == 0x88)
+  if (srcSize < 2 || srcData[0] == SPLIT_PATTERN &&
+          srcData[1] == SPLIT_PATTERN)
   {
-    #ifdef TEST
+    #ifdef WARNING
     *logofs << "UnpackJpeg: WARNING! Skipping unpack of dummy data.\n"
             << logofs_flush;
     #endif
@@ -147,7 +154,7 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
 
   #ifdef DEBUG
   *logofs << "UnpackJpeg: Decompression. Source size "
-          << src_size << " bits per plane " << dst_bpp
+          << srcSize << " bits per plane " << dstBpp
           << ".\n" << logofs_flush;
   #endif
 
@@ -155,9 +162,28 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
   srcGreenShift = ffs(geometry -> green_mask) - 1;
   srcBlueShift  = ffs(geometry -> blue_mask)  - 1;
 
-  srcRedMax     = geometry -> red_mask   >> srcRedShift;
-  srcGreenMax   = geometry -> green_mask >> srcGreenShift;
-  srcBlueMax    = geometry -> blue_mask  >> srcBlueShift;
+  #ifdef DEBUG
+  *logofs << "UnpackJpeg: Red shift " << (int) srcRedShift
+          << " green shift " << (int) srcGreenShift << " blue shift "
+          << (int) srcBlueShift << ".\n" << logofs_flush;
+  #endif
+
+  srcRedMax   = geometry -> red_mask   >> srcRedShift;
+  srcGreenMax = geometry -> green_mask >> srcGreenShift;
+  srcBlueMax  = geometry -> blue_mask  >> srcBlueShift;
+
+  #ifdef DEBUG
+  *logofs << "UnpackJpeg: Red mask " << (void *) geometry -> red_mask
+          << " green mask " << (void *) geometry -> green_mask
+          << " blue mask " << (void *) geometry -> blue_mask
+          << ".\n" << logofs_flush;
+  #endif
+
+  #ifdef DEBUG
+  *logofs << "UnpackJpeg: Red max " << srcRedMax << " green max "
+          << srcGreenMax << " blue max " << srcBlueMax
+          << ".\n" << logofs_flush;
+  #endif
 
   //
   // Make enough space in the temporary
@@ -165,14 +191,14 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
   // the image with 3 bytes for a pixel.
   //
 
-  tmpBufSize = dst_width * 3;
+  tmpBufSize = dstWidth * 3;
   tmpBuf     = new char[tmpBufSize];
 
   if (tmpBuf == NULL)
   {
     #ifdef PANIC
     *logofs << "UnpackJpeg: PANIC! Cannot allocate "
-            << dst_width * 3 << " bytes for Jpeg "
+            << dstWidth * 3 << " bytes for Jpeg "
             << "decompressed data.\n" << logofs_flush;
     #endif
 
@@ -183,53 +209,53 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
 
   int result = 1;
 
-  switch(dst_bpp)
+  switch(dstBpp)
   {
     case 8:
     {
       //
-      // Simply move the data from src_data to dst_data
+      // Simply move the data from srcData to dstData
       // taking into consideration the correct padding.
       //
 
       int row;
 
-      unsigned char * dstBuff = dst_data;
-      unsigned char * srcBuff = src_data;
+      unsigned char * dstBuff = dstData;
+      unsigned char * srcBuff = srcData;
 
-      for (row = 0; row < dst_height; row++)
+      for (row = 0; row < dstHeight; row++)
       {
-        memcpy(dstBuff, srcBuff, dst_width);
+        memcpy(dstBuff, srcBuff, dstWidth);
 
-        dstBuff += RoundUp4(dst_width);
-        srcBuff += dst_width;
+        dstBuff += RoundUp4(dstWidth);
+        srcBuff += dstWidth;
       }
 
       break;
     }
     case 16:
     {
-      result = DecompressJpeg16(src_data, src_size, dst_width,
-                                    dst_height, (unsigned short *) dst_data);
+      result = DecompressJpeg16(srcData, srcSize, dstWidth,
+                                    dstHeight, dstData, byteOrder);
       break;
     }
     case 24:
     {
-      result = DecompressJpeg24(src_data, src_size, dst_width,
-                                    dst_height, (unsigned char *) dst_data);
+      result = DecompressJpeg24(srcData, srcSize, dstWidth,
+                                    dstHeight, dstData, byteOrder);
       break;
     }
     case 32:
     {
-      result = DecompressJpeg32(src_data, src_size, dst_width,
-                                    dst_height, (unsigned int *) dst_data);
+      result = DecompressJpeg32(srcData, srcSize, dstWidth,
+                                    dstHeight, dstData, byteOrder);
       break;
     }
     default:
     {
       #ifdef PANIC
       *logofs << "UnpackJpeg: PANIC! Failed to decode Jpeg image. "
-              << " Unsupported Bpp value " << dst_bpp
+              << " Unsupported Bpp value " << dstBpp
               << " for the Jpeg compression"
               << ".\n" << logofs_flush;
       #endif
@@ -251,7 +277,7 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
 
     #ifdef PANIC
     *logofs << "UnpackJpeg: PANIC! Failed to decode Jpeg image using "
-            << dst_bpp << " Bpp destination.\n" << logofs_flush;
+            << dstBpp << " Bpp destination.\n" << logofs_flush;
     #endif
 
     return result;
@@ -333,15 +359,15 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
     }
   }
 
-  const T_colormask *colorMask = GetColorMask(maskMethod);
+  const T_colormask *colorMask = MethodColorMask(maskMethod);
 
-  unsigned char *dstBuff = dst_data;
+  unsigned char *dstBuff = dstData;
 
-  switch (dst_bpp)
+  switch (dstBpp)
   {
     case 16:
     {
-      Unpack16To16(colorMask, dstBuff, dstBuff, dstBuff + dst_size);
+      Unpack16To16(colorMask, dstBuff, dstBuff, dstBuff + dstSize);
 
       break;
     }
@@ -352,7 +378,7 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
     case 32:
     {
       Unpack32To32(colorMask, (unsigned int *) dstBuff, (unsigned int *) dstBuff,
-                       (unsigned int *) (dstBuff + dst_size));
+                       (unsigned int *) (dstBuff + dstSize));
       break;
     }
     default:
@@ -373,13 +399,15 @@ int UnpackJpeg(T_geometry *geometry, unsigned char method,
 //
 
 int DecompressJpeg16(unsigned char *compressedData, int compressedLen,
-                         unsigned int w, unsigned int h, unsigned short *dst_buf)
+                         unsigned int w, unsigned int h, unsigned char *dstBuf, int byteOrder)
 {
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
-  CARD16 *pixelPtr;
+  unsigned char *data;
   JSAMPROW rowPointer[1];
-  unsigned int dx, dy;
+
+  unsigned int dx = 0;
+  unsigned int dy = 0;
 
   #ifdef DEBUG
   *logofs << "DecompressJpeg16: Decompressing with length "
@@ -393,6 +421,16 @@ int DecompressJpeg16(unsigned char *compressedData, int compressedLen,
 
   jerr.error_exit = UnpackJpegErrorHandler;
 
+  if (setjmp(UnpackJpegContext) == 1)
+  {
+    #ifdef TEST
+    *logofs << "DecompressJpeg16: Out of the long jump with error '"
+            << jpegError << "'.\n" << logofs_flush;
+    #endif
+
+    goto AbortDecompressJpeg16;
+  }
+
   jpeg_create_decompress(&cinfo);
 
   if (jpegError) goto AbortDecompressJpeg16;
@@ -424,16 +462,16 @@ int DecompressJpeg16(unsigned char *compressedData, int compressedLen,
   }
 
   //
-  // PixelPtr points to dst_buf which is
+  // PixelPtr points to dstBuf which is
   // already padded correctly for the final
   // image to put
   //
 
-  pixelPtr = dst_buf;
+  data = dstBuf;
 
   rowPointer[0] = (JSAMPROW) tmpBuf;
 
-  dy = 0;
+  unsigned long pixel;
 
   while (cinfo.output_scanline < cinfo.output_height)
   {
@@ -443,16 +481,33 @@ int DecompressJpeg16(unsigned char *compressedData, int compressedLen,
 
     for (dx = 0; dx < w; dx++)
     {
-      *pixelPtr++ = RGB24_TO_PIXEL(16, tmpBuf[dx * 3], tmpBuf[dx * 3 + 1],
-                                       tmpBuf[dx * 3 + 2]);
+      pixel = RGB24_TO_PIXEL(16, tmpBuf[dx * 3], tmpBuf[dx * 3 + 1],
+                                 tmpBuf[dx * 3 + 2]);
+
+      //
+      // Follow the server byte order when arranging data.
+      //
+
+      if (byteOrder == 0)
+      {
+        data[0] = (unsigned char) (pixel & 0xff);
+        data[1] = (unsigned char) ((pixel >> 8) & 0xff);
+      }
+      else
+      {
+        data[1] = (unsigned char) (pixel & 0xff);
+        data[0] = (unsigned char) ((pixel >> 8) & 0xff);
+      }
+
+      data += 2;
     }
 
     //
-    // Move pixelPtr at the beginning of the
+    // Move data at the beginning of the
     // next line.
     //
 
-    (char *) pixelPtr += RoundUp4(w * 2) - w * 2;
+    data = data + (RoundUp4(w * 2) - w * 2);
 
     dy++;
   }
@@ -485,13 +540,15 @@ int DecompressJpeg16(unsigned char *compressedData, int compressedLen,
 }
 
 int DecompressJpeg24(unsigned char *compressedData, int compressedLen,
-                         unsigned int w, unsigned int h, unsigned char *dst_buf)
+                         unsigned int w, unsigned int h, unsigned char *dstBuf, int byteOrder)
 {
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   CARD8 *pixelPtr = NULL;
   JSAMPROW rowPointer[1];
-  unsigned int dx, dy;
+
+  unsigned int dx = 0;
+  unsigned int dy = 0;
 
   #ifdef TEST
   *logofs << "DecompressJpeg24: Decompressing with length "
@@ -505,6 +562,16 @@ int DecompressJpeg24(unsigned char *compressedData, int compressedLen,
 
   jerr.error_exit = UnpackJpegErrorHandler;
 
+  if (setjmp(UnpackJpegContext) == 1)
+  {
+    #ifdef TEST
+    *logofs << "DecompressJpeg24: Out of the long jump with error '"
+            << jpegError << "'.\n" << logofs_flush;
+    #endif
+
+    goto AbortDecompressJpeg24;
+  }
+
   jpeg_create_decompress(&cinfo);
 
   if (jpegError) goto AbortDecompressJpeg24;
@@ -536,17 +603,15 @@ int DecompressJpeg24(unsigned char *compressedData, int compressedLen,
   }
 
   //
-  // PixelPtr points to dst_buf which is
+  // PixelPtr points to dstBuf which is
   // already padded correctly for the final
   // image to put.
   //
 
-  pixelPtr = (CARD8 *) dst_buf;
+  pixelPtr = (CARD8 *) dstBuf;
 
   rowPointer[0] = (JSAMPROW) tmpBuf;
 
-  dy = 0;
-
   while (cinfo.output_scanline < cinfo.output_height)
   {
     jpeg_read_scanlines(&cinfo, rowPointer, 1);
@@ -555,14 +620,31 @@ int DecompressJpeg24(unsigned char *compressedData, int compressedLen,
 
     for (dx = 0; dx < w; dx++)
     {
-      *pixelPtr       = tmpBuf[dx * 3];
-      *(pixelPtr + 1) = tmpBuf[dx * 3 + 1];
-      *(pixelPtr + 2) = tmpBuf[dx * 3 + 2];
+      //
+      // Follow the server byte order when arranging data.
+      //
+
+      if (byteOrder == 0)
+      {
+        pixelPtr[0] = tmpBuf[dx * 3];
+        pixelPtr[1] = tmpBuf[dx * 3 + 1];
+        pixelPtr[2] = tmpBuf[dx * 3 + 2];
+      }
+      else
+      {
+        pixelPtr[2] = tmpBuf[dx * 3];
+        pixelPtr[1] = tmpBuf[dx * 3 + 1];
+        pixelPtr[0] = tmpBuf[dx * 3 + 2];
+      }
 
       pixelPtr += 3;
     }
 
-    (char *) pixelPtr += RoundUp4(w * 3) - w * 3;
+    //
+    // Go to the next line.
+    //
+
+    pixelPtr = (CARD8 *) (((char *) pixelPtr) + (RoundUp4(w * 3) - w * 3));
 
     dy++;
   }
@@ -595,13 +677,15 @@ int DecompressJpeg24(unsigned char *compressedData, int compressedLen,
 }
 
 int DecompressJpeg32(unsigned char *compressedData, int compressedLen,
-                         unsigned int w, unsigned int h, unsigned int *dst_buf)
+                         unsigned int w, unsigned int h, unsigned char *dstBuf, int byteOrder)
 {
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
-  CARD32 *pixelPtr = NULL;
+  unsigned char *data;
   JSAMPROW rowPointer[1];
-  unsigned int dx, dy;
+
+  unsigned int dx = 0;
+  unsigned int dy = 0;
 
   #ifdef TEST
   *logofs << "DecompressJpeg32: Decompressing with length "
@@ -615,6 +699,16 @@ int DecompressJpeg32(unsigned char *compressedData, int compressedLen,
 
   jerr.error_exit = UnpackJpegErrorHandler;
 
+  if (setjmp(UnpackJpegContext) == 1)
+  {
+    #ifdef TEST
+    *logofs << "DecompressJpeg32: Out of the long jump with error '"
+            << jpegError << "'.\n" << logofs_flush;
+    #endif
+
+    goto AbortDecompressJpeg32;
+  }
+
   jpeg_create_decompress(&cinfo);
 
   if (jpegError) goto AbortDecompressJpeg32;
@@ -646,16 +740,18 @@ int DecompressJpeg32(unsigned char *compressedData, int compressedLen,
   }
 
   //
-  // PixelPtr points to dst_buf which is
+  // PixelPtr points to dstBuf which is
   // already padded correctly for the final
   // image to put
   //
 
-  pixelPtr = (CARD32 *) dst_buf;
+  data = dstBuf;
 
   rowPointer[0] = (JSAMPROW) tmpBuf;
   
-  dy = 0;
+  unsigned long pixel;
+
+  int i;
 
   while (cinfo.output_scanline < cinfo.output_height)
   {
@@ -665,8 +761,31 @@ int DecompressJpeg32(unsigned char *compressedData, int compressedLen,
 
     for (dx = 0; dx < w; dx++)
     {
-      *pixelPtr++ = RGB24_TO_PIXEL(32, tmpBuf[dx * 3], tmpBuf[dx * 3 + 1],
-                                       tmpBuf[dx * 3 + 2]);
+      pixel = RGB24_TO_PIXEL(32, tmpBuf[dx * 3], tmpBuf[dx * 3 + 1],
+                                 tmpBuf[dx * 3 + 2]);
+
+      //
+      // Follow the server byte order when arranging data.
+      //
+
+      if (byteOrder == 0)
+      {
+        for (i = 0; i < 4; i++)
+        {
+          data[i] = (unsigned char)(pixel & 0xff);
+          pixel >>= 8;
+        }
+      }
+      else
+      {
+        for (i = 3; i >= 0; i--)
+        {
+          data[i] = (unsigned char) (pixel & 0xff);
+          pixel >>= 8;
+        }
+      }
+
+      data += 4;
     }
 
     dy++;
@@ -716,7 +835,7 @@ static boolean JpegFillInputBuffer(j_decompress_ptr cinfo)
 
 static void JpegSkipInputData(j_decompress_ptr cinfo, long num_bytes)
 {
-  if (num_bytes < 0 || (unsigned long)num_bytes > jpegSrcManager.bytes_in_buffer)
+  if (num_bytes < 0 || (unsigned long) num_bytes > jpegSrcManager.bytes_in_buffer)
   {
     jpegError = 1;
 
