@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -28,11 +28,12 @@
 #include "Types.h"
 #include "Timestamp.h"
 
-#include "Compressor.h"
-#include "Decompressor.h"
+#include "ActionCache.h"
 
-#include "StatusCache.h"
-#include "PositionCache.h"
+#include "ActionCacheCompat.h"
+#include "PositionCacheCompat.h"
+
+#include "StaticCompressor.h"
 
 //
 // Forward class declarations.
@@ -62,7 +63,10 @@ class WriteBuffer;
 #undef  REFERENCES
 
 //
-// Set default values.
+// Set default values.  We limit the maximum
+// size of a request to 262144 but we need to
+// consider the replies, whose size may be up
+// to 4MB.
 //
 
 #define MESSAGE_ENABLE_CACHE                        0
@@ -70,7 +74,7 @@ class WriteBuffer;
 #define MESSAGE_ENABLE_SPLIT                        0
 #define MESSAGE_ENABLE_COMPRESS                     0
 
-#define MESSAGE_DATA_LIMIT                          262144 - 4
+#define MESSAGE_DATA_LIMIT                          4194304 - 4
 #define MESSAGE_DATA_OFFSET                         4
 
 #define MESSAGE_CACHE_SLOTS                         6000
@@ -84,6 +88,7 @@ class WriteBuffer;
 class Message
 {
   friend class MessageStore;
+  friend class RenderExtensionStore;
 
   public:
 
@@ -116,6 +121,7 @@ class Message
   {
     size_   = message.size_;
     c_size_ = message.c_size_;
+    i_size_ = message.i_size_;
 
     hits_  = message.hits_;
     last_  = message.last_;
@@ -175,14 +181,6 @@ class Message
     #endif
   }
 
-  protected:
-
-  //
-  // This is the data part.
-  //
-
-  T_data data_;
-
   //
   // This is the original message size 
   // including the data part regardless
@@ -191,15 +189,29 @@ class Message
 
   int size_;
 
+  //
+  // This is the size of the identity.
+  //
+
+  int i_size_;
+
   //
   // This is the size, including identity,
   // after message has been 'updated' to 
   // reflect storage of data in compressed
-  // form.
+  // format.
   //
 
   int c_size_;
 
+  protected:
+
+  //
+  // This is the data part.
+  //
+
+  T_data data_;
+
   //
   // Time of last hit.
   //
@@ -251,7 +263,6 @@ class Message
 
 class MessageStore
 {
-
   public:
 
   //
@@ -317,29 +328,43 @@ class MessageStore
   // Last operation performed on cache.
   //
 
-  T_status    lastStatus;
-  StatusCache lastStatusCache;
+  T_store_action lastAction;
 
   //
   // Position of last element stored in cache.
   //
 
-  short int     lastAdded;
-  PositionCache lastAddedCache;
+  short int lastAdded;
 
   //
   // Positions of last element found in cache.
   //
 
-  short int     lastHit;
-  PositionCache lastHitCache;
+  short int lastHit;
 
   //
   // Position of last element erased.
   //
 
-  short int     lastRemoved;
-  PositionCache lastRemovedCache;
+  short int lastRemoved;
+
+  //
+  // Used to encode the the action to
+  // perform on the store and the slot
+  // involved.
+  //
+
+  ActionCache lastActionCache;
+
+  //
+  // Used in old protocol versions.
+  //
+
+  ActionCacheCompat lastActionCacheCompat;
+
+  PositionCacheCompat lastAddedCacheCompat;
+  PositionCacheCompat lastHitCacheCompat;
+  PositionCacheCompat lastRemovedCacheCompat;
 
   //
   // Position in cache where next insertion
@@ -349,8 +374,9 @@ class MessageStore
   short int lastRated;
 
   //
-  // Size of data part of last split 
-  // message once compressed.
+  // Size of data part of last split message
+  // once compressed. This is used only for
+  // compatibility with older proxies.
   //
 
   int lastResize;
@@ -361,8 +387,7 @@ class MessageStore
 
   public:
 
-  MessageStore(Compressor *compressor = NULL,
-                   Decompressor *decompressor = NULL);
+  MessageStore(StaticCompressor *compressor = NULL);
 
   virtual ~MessageStore();
 
@@ -384,9 +409,71 @@ class MessageStore
 
   virtual void destroy(Message *message) const = 0;
 
+  void validateSize(int size)
+  {
+    if (size < control -> MinimumMessageSize ||
+            size > control -> MaximumMessageSize)
+    {
+      *logofs << name() << ": PANIC! Invalid size " << size
+              << " for message.\n" << logofs_flush;
+
+      cerr << "Error" << ": Invalid size " << size
+           << " for message opcode " << opcode() << ".\n";
+
+      HandleAbort();
+    }
+  }
+
+  void validateSize(int dataSize, int compressedDataSize)
+  {
+    if (dataSize < 0 || dataSize > control ->
+            MaximumMessageSize - 4 || compressedDataSize < 0 ||
+                compressedDataSize >= dataSize)
+    {
+      *logofs << name() << ": PANIC! Invalid data size "
+              << dataSize << " and compressed data size "
+              << compressedDataSize << " for message.\n"
+              << logofs_flush;
+
+      cerr << "Error" << ": Invalid data size "
+           << dataSize << " and compressed data size "
+           << compressedDataSize << " for message "
+           << "opcode " << (unsigned) opcode() << ".\n";
+
+      HandleAbort();
+    }
+  }
+
+  //
+  // Determine if the message can be stored
+  // in the cache.
+  //
+
+  virtual int validateMessage(const unsigned char *buffer, int size)
+  {
+    return (size >= control -> MinimumMessageSize &&
+                size <= control -> MaximumMessageSize);
+  }
+
+  //
+  // Get data offset based on major and minor
+  // opcode of the message.
+  //
+
+  virtual int identitySize(const unsigned char *buffer, unsigned int size)
+  {
+    return dataOffset;
+  }
+
   //
   // Encode identity and data using the
-  // best message's specific encoding.
+  // specific message encoding.
+  //
+  // Some messages do not implement these
+  // methods because the encoding is done
+  // directly in the channel loop. Should
+  // move the encoding methods in in the
+  // message classes.
   //
 
   virtual int encodeIdentity(EncodeBuffer &encodeBuffer, const unsigned char *buffer,
@@ -424,14 +511,18 @@ class MessageStore
   }
 
   //
-  // Post process message contained in store
-  // updating the data part once it has been
-  // completely transferred at the peer side.
+  // Post process the message information
+  // contained in the store by either up-
+  // dating the size record or the actual
+  // data part once the message has been
+  // completely sent to our peer.
   //
 
   void updateData(const int position, unsigned int dataSize,
                       unsigned int compressedDataSize);
 
+  void updateData(const T_checksum checksum, unsigned int compressedDataSize);
+
   void updateData(const int position, const unsigned char *newData,
                       unsigned int dataSize, unsigned int compressedDataSize);
 
@@ -454,32 +545,14 @@ class MessageStore
 
   virtual void dumpIdentity(const Message *message) const = 0;
 
-  //
-  // These are called before and after the parsing
-  // procedure. They can be used to modify parts of
-  // the target message. These functions are made
-  // outdated by step 3 of the protocol.
-  //
-
-  virtual int parseBegin(Message *message, unsigned char *buffer,
-                             unsigned int size, int bigEndian) const
-  {
-    return 1;
-  }
-
-  virtual int parseEnd(Message *message, unsigned char *buffer,
-                           unsigned int size, int bigEndian) const
-  {
-    return 1;
-  }
-
   //
   // Design should preserve these from being
   // virtual.
   //
 
-  int parseData(Message *message, const unsigned char *buffer, unsigned int size,
-                    T_checksum_action checksumAction, T_data_action dataAction, int bigEndian);
+  int parseData(Message *message, int split, const unsigned char *buffer,
+                    unsigned int size, T_checksum_action checksumAction,
+                        T_data_action dataAction, int bigEndian);
 
   int parseData(Message *message, const unsigned char *buffer,
                     unsigned int size, const unsigned char *compressedData,
@@ -489,6 +562,11 @@ class MessageStore
   int unparseData(const Message *message, unsigned char *buffer,
                       unsigned int size, int bigEndian);
 
+  //
+  // Manage efficient allocation of messages
+  // in the heap.
+  //
+
   void recycle(Message *message)
   {
     #ifdef TEST
@@ -566,16 +644,16 @@ class MessageStore
                         unsigned int size, int bigEndian)
   {
     //
-    // Messages that have data part starting at
-    // an offset possibly beyond the end of the
-    // message must include size in identity
-    // checksum.
+    // Messages that have a data part starting
+    // at an offset possibly beyond the end of
+    // the message, must include the size in
+    // the identity checksum.
     //
 
-    if ((int) size > dataOffset)
+    if ((int) size > message -> i_size_)
     {
-      md5_append(md5_state_, buffer + dataOffset,
-                     size - dataOffset);
+      md5_append(md5_state_, buffer + message -> i_size_,
+                     size - message -> i_size_);
     }
   }
 
@@ -591,7 +669,7 @@ class MessageStore
   // time of data parsing.
   //
 
-  int parse(Message *message, const unsigned char *buffer, unsigned int size,
+  int parse(Message *message, int split, const unsigned char *buffer, unsigned int size,
                 T_checksum_action checksumAction, T_data_action dataAction, int bigEndian);
 
   int parse(Message *message, const unsigned char *buffer, unsigned int size,
@@ -610,19 +688,6 @@ class MessageStore
                 unparseIdentity(message, buffer, size, bigEndian));
   }
 
-  //
-  // This function is new. It leverages the
-  // fact that data part is always put at
-  // the offset set in the message class.
-  //
-
-  int unparse(const Message *message, unsigned char *identity,
-                  unsigned char *data, unsigned int size, int bigEndian)
-  {
-    return (unparseData(message, data - dataOffset, size, bigEndian) &&
-                unparseIdentity(message, identity, size, bigEndian));
-  }
-
   void dump(const Message *message) const
   {
     dumpIdentity(message);
@@ -663,19 +728,19 @@ class MessageStore
 
   Message *get(const int position) const
   {
-    if (position >= (int) messages_ -> size() || position < 0)
+    if (position < 0 || position >= cacheSlots)
     {
       #ifdef PANIC
       *logofs << name() << ": PANIC! Requested position "
-              << position << " is beyond the end of "
+              << position << " is not inside the "
               << "container.\n" << logofs_flush;
       #endif
 
       cerr << "Error" << ": Requested position "
-           << position << " is beyond the end of "
+           << position << " is not inside the"
            << "container.\n";
 
-      return NULL;
+      HandleAbort();
     }
     else if ((*messages_)[position] == NULL)
     {
@@ -687,34 +752,20 @@ class MessageStore
 
       cerr << "Error" << ": Message at position "
            << position << " is NULL.\n";
+
+      HandleAbort();
     }
 
     return (*messages_)[position];
   }
 
-  //
-  // Find a message in cache. Returns the 
-  // position, if found, and save iterator
-  // into the provided reference.
-  //
-
-  int find(const Message *message, 
-               T_checksums::iterator &found) const;
-
-  int find(const Message *message) const
-  {
-    T_checksums::iterator found;
-
-    return find(message, found);
-  }
-
   //
   // This is the method called at encoding
   // side to add a message to cache.
   //
 
   int findOrAdd(Message *message, T_checksum_action checksumAction,
-                    int &added, int &locked);
+                    T_data_action dataAction, int &added, int &locked);
 
   //
   // Utility interfaces to message insertion
@@ -722,9 +773,10 @@ class MessageStore
   //
 
   int add(Message *message, const int position,
-              T_checksum_action checksumAction);
+              T_checksum_action checksumAction, T_data_action dataAction);
 
-  int remove(const int position, T_checksum_action checksumAction);
+  int remove(const int position, T_checksum_action checksumAction,
+                 T_data_action dataAction);
 
   //
   // Make space in the repository by remove
@@ -734,70 +786,12 @@ class MessageStore
   int clean(T_checksum_action checksumAction);
 
   //
-  // Increase or decrease "rate" of objects.
+  // Increase or decrease the "rating" of
+  // the message object.
   //
 
-  int touch(const int position) const
-  {
-    Message *message = (*messages_)[position];
-
-    return touch(message);
-  }
-
-  int untouch(const int position) const
-  {
-    Message *message = (*messages_)[position];
-
-    return untouch(message);
-  }
-
-  int touch(Message *message) const
-  {
-    if (message == NULL)
-    {
-      return -1;
-    }
-
-    message -> last_ = timestamp_.tv_sec;
-
-    message -> hits_ += control -> StoreHitsTouch;
-
-    if (message -> hits_ > control -> StoreHitsLimit)
-    {
-      message -> hits_ = control -> StoreHitsLimit;
-    }
-
-    #ifdef TEST
-    *logofs << name() << ": Increased hits of object to "
-            << message -> hits_ <<  " at " << strTimestamp(timestamp_)
-            << ".\n" << logofs_flush;
-    #endif
-
-    return message -> hits_;
-  }
-
-  int untouch(Message *message) const
-  {
-    if (message == NULL)
-    {
-      return -1;
-    }
-
-    message -> hits_ -= control -> StoreHitsUntouch;
-
-    if (message -> hits_ < 0)
-    {
-      message -> hits_ = 0;
-    }
-
-    #ifdef TEST
-    *logofs << name() << ": Decreased hits of object to " 
-            << message -> hits_ << ".\n"
-            << logofs_flush;
-    #endif
-
-    return message -> hits_;
-  }
+  int touch(Message *message) const;
+  int untouch(Message *message) const;
 
   int getTouches(const int position) const
   {
@@ -825,7 +819,7 @@ class MessageStore
   // ry to keep object in cache.
   //
 
-  int getRate(const int position, T_rate type) const;
+  int getRating(Message *message, T_rating type) const;
 
   //
   // Increase or decrease locks of message at given
@@ -849,31 +843,23 @@ class MessageStore
     return message -> locks_;
   }
 
-  md5_byte_t * const getChecksum(const int position) const
+  T_checksum const getChecksum(const int position) const
   {
-    Message *message = get(position);
-
-    if (message != NULL)
-    {
-      return getChecksum(message);
-    }
-    else
-    {
-      return NULL;
-    }
+    return getChecksum(get(position));
   }
 
-  md5_byte_t * const getChecksum(const Message *message) const
+  T_checksum const getChecksum(const Message *message) const
   {
     if (message -> md5_digest_ == NULL)
     {
       #ifdef PANIC
-      *logofs << name() << ": PANIC! Checksum not initialized for object at " 
-              << message << ".\n" << logofs_flush;
+      *logofs << name() << ": PANIC! Checksum not initialized "
+              << "for object at " << message << ".\n"
+              << logofs_flush;
       #endif
 
-      cerr << "Error" << ": Checksum not initialized for object at "
-           << message << ".\n";
+      cerr << "Error" << ": Checksum not initialized "
+           << "for object at " << message << ".\n";
 
       HandleAbort();
     }
@@ -886,27 +872,35 @@ class MessageStore
     return message -> md5_digest_;
   }
 
-  int getSize(const int position) const
+  //
+  // Calculate the checksum on the fly based the
+  // opcode in the buffer. Useful in the case a
+  // message was not processed or was not stored
+  // in the cache. The returned checksum must be
+  // explicitly deallocated by the caller, after
+  // use.
+  //
+
+  T_checksum getChecksum(const unsigned char *buffer,
+                             unsigned int size, int bigEndian);
+
+  const unsigned char *getData(const Message *message) const
   {
-    Message *message = (*messages_)[position];
-
-    if (message == NULL)
-    {
-      return 0;
-    }
+    return message -> data_.begin();
+  }
 
+  int plainSize(const Message *message) const
+  {
     return message -> size_;
   }
 
-  int getCSize(const int position) const
+  int identitySize(Message *message)
   {
-    Message *message = (*messages_)[position];
-
-    if (message == NULL)
-    {
-      return 0;
-    }
+    return message -> i_size_;
+  }
 
+  int compressedSize(const Message *message) const
+  {
     return message -> c_size_;
   }
 
@@ -925,11 +919,6 @@ class MessageStore
     temporary_ = NULL;
   }
 
-  void newTimestamp()
-  {
-    timestamp_ = getTimestamp();
-  }
-
   //
   // On side where we don't have checksums, we
   // count how many messages are in the array.
@@ -983,7 +972,7 @@ class MessageStore
 
   int saveStore(ostream *cachefs, md5_state_t *md5_state_stream,
                     md5_state_t *md5_state_client, T_checksum_action checksumAction,
-                        T_data_action dataAction, int bigEndian) const;
+                        T_data_action dataAction, int bigEndian);
 
   int loadStore(istream *cachefs, md5_state_t *md5_state_stream,
                     T_checksum_action checksumAction, T_data_action dataAction,
@@ -1030,12 +1019,6 @@ class MessageStore
 
   md5_state_t *md5_state_;
 
-  //
-  // Timestamp of last operation on store.
-  //
-
-  T_timestamp timestamp_;
-
   //
   // How many messages have been marked
   // to be retained in repository.
@@ -1049,8 +1032,7 @@ class MessageStore
   // Used to compress data payload.
   //
 
-  Compressor   *compressor_;
-  Decompressor *decompressor_;
+  StaticCompressor *compressor_;
 
   //
   // Keep track of how many bytes 
@@ -1074,5 +1056,71 @@ class MessageStore
   #endif
 };
 
+//
+// This is an ancillary class of the message
+// store, used to encode extensions based on
+// the minor opcode.
+//
+
+class MinorMessageStore
+{
+  public:
+
+  virtual ~MinorMessageStore()
+  {
+  }
+
+  virtual const char *name() const = 0;
+
+  virtual int identitySize(const unsigned char *buffer, unsigned int size) = 0;
+
+  virtual int encodeMessage(EncodeBuffer &encodeBuffer, const unsigned char *buffer,
+                                const unsigned int size, int bigEndian,
+                                    ChannelCache *channelCache) const
+  {
+    return 1;
+  }
+
+  virtual int decodeMessage(DecodeBuffer &decodeBuffer, unsigned char *&buffer,
+                                unsigned int &size, unsigned char type, int bigEndian,
+                                    WriteBuffer *writeBuffer, ChannelCache *channelCache) const
+  {
+    return 1;
+  }
+
+  virtual void encodeData(EncodeBuffer &encodeBuffer, const unsigned char *buffer,
+                              unsigned int size, int bigEndian,
+                                  ChannelCache *channelCache) const
+  {
+  }
+
+  virtual void decodeData(DecodeBuffer &decodeBuffer, unsigned char *buffer,
+                              unsigned int size, int bigEndian,
+                                  ChannelCache *channelCache) const
+  {
+  }
+
+  virtual int parseIdentity(Message *message, const unsigned char *buffer,
+                                unsigned int size, int bigEndian) const = 0;
+
+  virtual int unparseIdentity(const Message *message, unsigned char *buffer,
+                                  unsigned int size, int bigEndian) const = 0;
+
+  virtual void updateIdentity(EncodeBuffer &encodeBuffer, const Message *message,
+                                  const Message *cachedMessage,
+                                      ChannelCache *channelCache) const
+  {
+  }
+
+  virtual void updateIdentity(DecodeBuffer &decodeBuffer, const Message *message,
+                                  ChannelCache *channelCache) const
+  {
+  }
+
+  virtual void identityChecksum(const Message *message, const unsigned char *buffer,
+                                    unsigned int size, md5_state_t *md5_state,
+                                        int bigEndian) const = 0;
+};
+
 #endif /* Message_H */
 
