@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -38,22 +38,14 @@
 // Constructors and destructors.
 //
 
-PutPackedImageStore::PutPackedImageStore(Compressor *compressor, Decompressor *decompressor)
+PutPackedImageStore::PutPackedImageStore(StaticCompressor *compressor)
 
-  : MessageStore(compressor, decompressor)
+  : MessageStore(compressor)
 {
   enableCache    = PUTPACKEDIMAGE_ENABLE_CACHE;
   enableData     = PUTPACKEDIMAGE_ENABLE_DATA;
   enableSplit    = PUTPACKEDIMAGE_ENABLE_SPLIT;
-
-  if (control -> isProtoStep3() == 1)
-  {
-    enableCompress = PUTPACKEDIMAGE_ENABLE_COMPRESS_IF_PROTO_STEP_3;
-  }
-  else
-  {
-    enableCompress = PUTPACKEDIMAGE_ENABLE_COMPRESS;
-  }
+  enableCompress = PUTPACKEDIMAGE_ENABLE_COMPRESS;
 
   dataLimit  = PUTPACKEDIMAGE_DATA_LIMIT;
   dataOffset = PUTPACKEDIMAGE_DATA_OFFSET;
@@ -98,19 +90,9 @@ int PutPackedImageStore::encodeIdentity(EncodeBuffer &encodeBuffer, const unsign
   *logofs << name() << ": Encoding full message identity.\n" << logofs_flush;
   #endif
 
-  //
-  // TODO: Compression of this message must
-  // carefully reviewed. Probably checksum
-  // is sent to let the remote peer abort
-  // the split even when it is not needed.
-  //
-
-  if (control -> isProtoStep4() == 1)
-  {
-    // Client.
-    encodeBuffer.encodeCachedValue(*(buffer + 1), 8,
-                       clientCache -> resourceCache);
-  }
+  // Client.
+  encodeBuffer.encodeCachedValue(*(buffer + 1), 8,
+                     clientCache -> resourceCache);
 
   // Size.
   encodeBuffer.encodeValue(GetUINT(buffer + 2, bigEndian), 16, 10);
@@ -123,7 +105,7 @@ int PutPackedImageStore::encodeIdentity(EncodeBuffer &encodeBuffer, const unsign
                      clientCache -> gcCache);
   // Method.
   encodeBuffer.encodeCachedValue(buffer[12], 8,
-                     clientCache -> putPackedImageMethodCache);
+                     clientCache -> methodCache);
   // Format.
   encodeBuffer.encodeValue(buffer[13], 2);
 
@@ -196,12 +178,9 @@ int PutPackedImageStore::decodeIdentity(DecodeBuffer &decodeBuffer, unsigned cha
   unsigned int value;
   unsigned char cValue;
 
-  if (control -> isProtoStep4() == 1)
-  {
-    // Client.
-    decodeBuffer.decodeCachedValue(cValue, 8,
-                       clientCache -> resourceCache);
-  }
+  // Client.
+  decodeBuffer.decodeCachedValue(cValue, 8,
+                     clientCache -> resourceCache);
 
   // Size.
   decodeBuffer.decodeValue(size, 16, 10);
@@ -222,7 +201,7 @@ int PutPackedImageStore::decodeIdentity(DecodeBuffer &decodeBuffer, unsigned cha
 
   // Method.
   decodeBuffer.decodeCachedValue(cValue, 8,
-                     clientCache -> putPackedImageMethodCache);
+                     clientCache -> methodCache);
   *(buffer + 12) = cValue;
 
   // Format.
@@ -313,10 +292,7 @@ int PutPackedImageStore::parseIdentity(Message *message, const unsigned char *bu
   // Here is the fingerprint.
   //
 
-  if (control -> isProtoStep4() == 1)
-  {
-    putPackedImage -> client = *(buffer + 1);
-  }
+  putPackedImage -> client = *(buffer + 1);
 
   putPackedImage -> drawable = GetULONG(buffer + 4,  bigEndian); 
   putPackedImage -> gcontext = GetULONG(buffer + 8,  bigEndian);
@@ -357,10 +333,7 @@ int PutPackedImageStore::unparseIdentity(const Message *message, unsigned char *
   // Fill all the message's fields.
   //
 
-  if (control -> isProtoStep4() == 1)
-  {
-    *(buffer + 1) = putPackedImage -> client;
-  }
+  *(buffer + 1) = putPackedImage -> client;
 
   PutULONG(putPackedImage -> drawable, buffer + 4, bigEndian);
   PutULONG(putPackedImage -> gcontext, buffer + 8, bigEndian);
@@ -432,26 +405,17 @@ void PutPackedImageStore::identityChecksum(const Message *message, const unsigne
                                                unsigned int size, int bigEndian) const
 {
   //
-  // Fields format, src_depth, dst_depth, src_length,
-  // dst_length, src_x, src_y, src_width, src_height.
-  // Method was encoded in differences in protocol
-  // versions before 3.
+  // Fields method, format, src_depth, dst_depth,
+  // src_length, dst_length, src_x, src_y, src_width,
+  // src_height.
+  //
+  //
+  // TODO: We should better investigate the effect of
+  // having fields src_x and src_y in identity instead
+  // of keeping them in differences.
   //
 
-  if (control -> isProtoStep3() == 1)
-  {
-    //
-    // TODO: We should better investigate the effect
-    // of having fields src_x and src_y in identity
-    // instead of in differences.
-    //
-
-    md5_append(md5_state_, buffer + 12, 20);
-  }
-  else
-  {
-    md5_append(md5_state_, buffer + 13, 19);
-  }
+  md5_append(md5_state_, buffer + 12, 20);
 }
 
 void PutPackedImageStore::updateIdentity(EncodeBuffer &encodeBuffer, const Message *message,
@@ -467,40 +431,23 @@ void PutPackedImageStore::updateIdentity(EncodeBuffer &encodeBuffer, const Messa
 
   ClientCache *clientCache = (ClientCache *) channelCache;
 
-  if (control -> isProtoStep4() == 1)
-  {
-    #ifdef TEST
-    *logofs << name() << ": Encoding value "
-            << (unsigned int) putPackedImage -> client
-            << " as client field.\n" << logofs_flush;
-    #endif
+  #ifdef TEST
+  *logofs << name() << ": Encoding value "
+          << (unsigned int) putPackedImage -> client
+          << " as client field.\n" << logofs_flush;
+  #endif
 
-    encodeBuffer.encodeCachedValue(putPackedImage -> client, 8,
-                       clientCache -> resourceCache);
+  encodeBuffer.encodeCachedValue(putPackedImage -> client, 8,
+                     clientCache -> resourceCache);
 
-    cachedPutPackedImage -> client = putPackedImage -> client;
-  }
-
-  if (control -> isProtoStep3() == 0)
-  {
-    #ifdef TEST
-    *logofs << name() << ": Encoding value " << putPackedImage -> method
-            << " as method field.\n" << logofs_flush;
-    #endif
-
-    encodeBuffer.encodeCachedValue(putPackedImage -> method, 8,
-                       clientCache -> putPackedImageMethodCache);
-
-    cachedPutPackedImage -> method = putPackedImage -> method;
-  }
+  cachedPutPackedImage -> client = putPackedImage -> client;
 
   #ifdef TEST
   *logofs << name() << ": Encoding value " << putPackedImage -> drawable
           << " as drawable field.\n" << logofs_flush;
   #endif
 
-  encodeBuffer.encodeXidValue(putPackedImage -> drawable,
-                     cachedPutPackedImage -> drawable, clientCache -> drawableCache);
+  encodeBuffer.encodeXidValue(putPackedImage -> drawable, clientCache -> drawableCache);
 
   cachedPutPackedImage -> drawable = putPackedImage -> drawable;
 
@@ -509,8 +456,7 @@ void PutPackedImageStore::updateIdentity(EncodeBuffer &encodeBuffer, const Messa
           << " as gcontext field.\n" << logofs_flush;
   #endif
 
-  encodeBuffer.encodeXidValue(putPackedImage -> gcontext,
-                     cachedPutPackedImage -> gcontext, clientCache -> gcCache);
+  encodeBuffer.encodeXidValue(putPackedImage -> gcontext, clientCache -> gcCache);
 
   cachedPutPackedImage -> gcontext = putPackedImage -> gcontext;
 
@@ -571,35 +517,17 @@ void PutPackedImageStore::updateIdentity(DecodeBuffer &decodeBuffer, const Messa
   ClientCache *clientCache = (ClientCache *) channelCache;
 
   unsigned int  value;
-  unsigned char cValue;
 
-  if (control -> isProtoStep4() == 1)
-  {
-    decodeBuffer.decodeCachedValue(putPackedImage -> client, 8,
-                       clientCache -> resourceCache);
+  decodeBuffer.decodeCachedValue(putPackedImage -> client, 8,
+                     clientCache -> resourceCache);
 
-    #ifdef DEBUG
-    *logofs << name() << ": Decoded value "
-            << (unsigned int) putPackedImage -> client
-            << " as client field.\n" << logofs_flush;
-    #endif
-  }
+  #ifdef DEBUG
+  *logofs << name() << ": Decoded value "
+          << (unsigned int) putPackedImage -> client
+          << " as client field.\n" << logofs_flush;
+  #endif
 
-  if (control -> isProtoStep3() == 0)
-  {
-    decodeBuffer.decodeCachedValue(cValue, 8,
-                       clientCache -> putPackedImageMethodCache);
-
-    putPackedImage -> method = cValue;
-
-    #ifdef DEBUG
-    *logofs << name() << ": Decoded value " << putPackedImage -> method
-            << " as method field.\n" << logofs_flush;
-    #endif
-  }
-
-  decodeBuffer.decodeXidValue(value, putPackedImage -> drawable,
-                     clientCache -> drawableCache);
+  decodeBuffer.decodeXidValue(value, clientCache -> drawableCache);
 
   putPackedImage -> drawable = value;
 
@@ -608,8 +536,7 @@ void PutPackedImageStore::updateIdentity(DecodeBuffer &decodeBuffer, const Messa
           << " as drawable field.\n" << logofs_flush;
   #endif
 
-  decodeBuffer.decodeXidValue(value, putPackedImage -> gcontext,
-                     clientCache -> gcCache);
+  decodeBuffer.decodeXidValue(value, clientCache -> gcCache);
 
   putPackedImage -> gcontext = value;
 
