@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -37,6 +37,9 @@
 // Log level.
 //
 
+#define PANIC
+#define WARNING
+#undef  TEST
 #undef  DEBUG
 
 //
@@ -54,8 +57,9 @@ Statistics::Statistics(Proxy *proxy) : proxy_(proxy)
   transportPartial_.writeTime_            = 0;
   transportPartial_.proxyBytesIn_         = 0;
   transportPartial_.proxyBytesOut_        = 0;
-  transportPartial_.proxyFrameIn_         = 0;
-  transportPartial_.proxyFrameOut_        = 0;
+  transportPartial_.proxyFramesIn_        = 0;
+  transportPartial_.proxyFramesOut_       = 0;
+  transportPartial_.proxyWritesOut_       = 0;
   transportPartial_.compressedBytesIn_    = 0;
   transportPartial_.compressedBytesOut_   = 0;
   transportPartial_.decompressedBytesIn_  = 0;
@@ -67,8 +71,9 @@ Statistics::Statistics(Proxy *proxy) : proxy_(proxy)
   transportTotal_.writeTime_            = 0;
   transportTotal_.proxyBytesIn_         = 0;
   transportTotal_.proxyBytesOut_        = 0;
-  transportTotal_.proxyFrameIn_         = 0;
-  transportTotal_.proxyFrameOut_        = 0;
+  transportTotal_.proxyFramesIn_        = 0;
+  transportTotal_.proxyFramesOut_       = 0;
+  transportTotal_.proxyWritesOut_       = 0;
   transportTotal_.compressedBytesIn_    = 0;
   transportTotal_.compressedBytesOut_   = 0;
   transportTotal_.decompressedBytesIn_  = 0;
@@ -83,6 +88,11 @@ Statistics::Statistics(Proxy *proxy) : proxy_(proxy)
     protocolPartial_.requestBitsIn_[i]  = 0;
     protocolPartial_.requestBitsOut_[i] = 0;
 
+    protocolPartial_.renderRequestCached_[i]  = 0;
+    protocolPartial_.renderRequestCount_[i]   = 0;
+    protocolPartial_.renderRequestBitsIn_[i]  = 0;
+    protocolPartial_.renderRequestBitsOut_[i] = 0;
+
     protocolPartial_.replyCached_[i]  = 0;
     protocolPartial_.replyCount_[i]   = 0;
     protocolPartial_.replyBitsIn_[i]  = 0;
@@ -99,6 +109,11 @@ Statistics::Statistics(Proxy *proxy) : proxy_(proxy)
     protocolTotal_.requestBitsIn_[i]  = 0;
     protocolTotal_.requestBitsOut_[i] = 0;
 
+    protocolTotal_.renderRequestCached_[i]  = 0;
+    protocolTotal_.renderRequestCount_[i]   = 0;
+    protocolTotal_.renderRequestBitsIn_[i]  = 0;
+    protocolTotal_.renderRequestBitsOut_[i] = 0;
+
     protocolTotal_.replyCached_[i]  = 0;
     protocolTotal_.replyCount_[i]   = 0;
     protocolTotal_.replyBitsIn_[i]  = 0;
@@ -110,37 +125,53 @@ Statistics::Statistics(Proxy *proxy) : proxy_(proxy)
     protocolTotal_.eventBitsOut_[i] = 0;
   }
 
-  protocolPartial_.syncCount_   = 0;
-  protocolPartial_.syncBitsIn_  = 0;
-  protocolPartial_.syncBitsOut_ = 0;
+  protocolPartial_.cupsCount_    = 0;
+  protocolPartial_.cupsBitsIn_   = 0;
+  protocolPartial_.cupsBitsOut_  = 0;
 
-  protocolPartial_.keybdCount_   = 0;
-  protocolPartial_.keybdBitsIn_  = 0;
-  protocolPartial_.keybdBitsOut_ = 0;
-
-  protocolPartial_.sambaCount_   = 0;
-  protocolPartial_.sambaBitsIn_  = 0;
-  protocolPartial_.sambaBitsOut_ = 0;
+  protocolPartial_.smbCount_     = 0;
+  protocolPartial_.smbBitsIn_    = 0;
+  protocolPartial_.smbBitsOut_   = 0;
 
   protocolPartial_.mediaCount_   = 0;
   protocolPartial_.mediaBitsIn_  = 0;
   protocolPartial_.mediaBitsOut_ = 0;
 
-  protocolTotal_.syncCount_     = 0;
-  protocolTotal_.syncBitsIn_    = 0;
-  protocolTotal_.syncBitsOut_   = 0;
+  protocolPartial_.httpCount_    = 0;
+  protocolPartial_.httpBitsIn_   = 0;
+  protocolPartial_.httpBitsOut_  = 0;
 
-  protocolTotal_.keybdCount_     = 0;
-  protocolTotal_.keybdBitsIn_    = 0;
-  protocolTotal_.keybdBitsOut_   = 0;
+  protocolPartial_.fontCount_    = 0;
+  protocolPartial_.fontBitsIn_   = 0;
+  protocolPartial_.fontBitsOut_  = 0;
 
-  protocolTotal_.sambaCount_     = 0;
-  protocolTotal_.sambaBitsIn_    = 0;
-  protocolTotal_.sambaBitsOut_   = 0;
+  protocolPartial_.slaveCount_   = 0;
+  protocolPartial_.slaveBitsIn_  = 0;
+  protocolPartial_.slaveBitsOut_ = 0;
 
-  protocolTotal_.mediaCount_     = 0;
-  protocolTotal_.mediaBitsIn_    = 0;
-  protocolTotal_.mediaBitsOut_   = 0;
+  protocolTotal_.cupsCount_    = 0;
+  protocolTotal_.cupsBitsIn_   = 0;
+  protocolTotal_.cupsBitsOut_  = 0;
+
+  protocolTotal_.smbCount_     = 0;
+  protocolTotal_.smbBitsIn_    = 0;
+  protocolTotal_.smbBitsOut_   = 0;
+
+  protocolTotal_.mediaCount_   = 0;
+  protocolTotal_.mediaBitsIn_  = 0;
+  protocolTotal_.mediaBitsOut_ = 0;
+
+  protocolTotal_.httpCount_    = 0;
+  protocolTotal_.httpBitsIn_   = 0;
+  protocolTotal_.httpBitsOut_  = 0;
+
+  protocolTotal_.fontCount_    = 0;
+  protocolTotal_.fontBitsIn_   = 0;
+  protocolTotal_.fontBitsOut_  = 0;
+
+  protocolTotal_.slaveCount_   = 0;
+  protocolTotal_.slaveBitsIn_  = 0;
+  protocolTotal_.slaveBitsOut_ = 0;
 
   packedPartial_.packedBytesIn_  = 0;
   packedPartial_.packedBytesOut_ = 0;
@@ -161,6 +192,13 @@ Statistics::Statistics(Proxy *proxy) : proxy_(proxy)
 
   overallTotal_.overallBytesIn_    = 0;
   overallTotal_.overallBytesOut_   = 0;
+
+  proxyData_.protocolCount_ = 0;
+  proxyData_.controlCount_  = 0;
+  proxyData_.splitCount_    = 0;
+  proxyData_.dataCount_     = 0;
+
+  proxyData_.streamRatio_ = 1;
 }
 
 Statistics::~Statistics()
@@ -174,8 +212,9 @@ int Statistics::resetPartialStats()
   transportPartial_.writeTime_            = 0;
   transportPartial_.proxyBytesIn_         = 0;
   transportPartial_.proxyBytesOut_        = 0;
-  transportPartial_.proxyFrameIn_         = 0;
-  transportPartial_.proxyFrameOut_        = 0;
+  transportPartial_.proxyFramesIn_        = 0;
+  transportPartial_.proxyFramesOut_       = 0;
+  transportPartial_.proxyWritesOut_       = 0;
   transportPartial_.compressedBytesIn_    = 0;
   transportPartial_.compressedBytesOut_   = 0;
   transportPartial_.decompressedBytesIn_  = 0;
@@ -190,6 +229,11 @@ int Statistics::resetPartialStats()
     protocolPartial_.requestBitsIn_[i]  = 0;
     protocolPartial_.requestBitsOut_[i] = 0;
 
+    protocolPartial_.renderRequestCached_[i]  = 0;
+    protocolPartial_.renderRequestCount_[i]   = 0;
+    protocolPartial_.renderRequestBitsIn_[i]  = 0;
+    protocolPartial_.renderRequestBitsOut_[i] = 0;
+
     protocolPartial_.replyCached_[i]  = 0;
     protocolPartial_.replyCount_[i]   = 0;
     protocolPartial_.replyBitsIn_[i]  = 0;
@@ -201,22 +245,30 @@ int Statistics::resetPartialStats()
     protocolPartial_.eventBitsOut_[i] = 0;
   }
 
-  protocolPartial_.syncCount_   = 0;
-  protocolPartial_.syncBitsIn_  = 0;
-  protocolPartial_.syncBitsOut_ = 0;
+  protocolPartial_.cupsCount_    = 0;
+  protocolPartial_.cupsBitsIn_   = 0;
+  protocolPartial_.cupsBitsOut_  = 0;
 
-  protocolPartial_.keybdCount_   = 0;
-  protocolPartial_.keybdBitsIn_  = 0;
-  protocolPartial_.keybdBitsOut_ = 0;
-
-  protocolPartial_.sambaCount_   = 0;
-  protocolPartial_.sambaBitsIn_  = 0;
-  protocolPartial_.sambaBitsOut_ = 0;
+  protocolPartial_.smbCount_     = 0;
+  protocolPartial_.smbBitsIn_    = 0;
+  protocolPartial_.smbBitsOut_   = 0;
 
   protocolPartial_.mediaCount_   = 0;
   protocolPartial_.mediaBitsIn_  = 0;
   protocolPartial_.mediaBitsOut_ = 0;
 
+  protocolPartial_.httpCount_    = 0;
+  protocolPartial_.httpBitsIn_   = 0;
+  protocolPartial_.httpBitsOut_  = 0;
+
+  protocolPartial_.fontCount_    = 0;
+  protocolPartial_.fontBitsIn_   = 0;
+  protocolPartial_.fontBitsOut_  = 0;
+
+  protocolPartial_.slaveCount_   = 0;
+  protocolPartial_.slaveBitsIn_  = 0;
+  protocolPartial_.slaveBitsOut_ = 0;
+
   packedPartial_.packedBytesIn_  = 0;
   packedPartial_.packedBytesOut_ = 0;
 
@@ -256,9 +308,9 @@ int Statistics::getClientCacheStats(int type, char *&buffer)
   strcat(buffer, "\nNX Cache Statistics\n");
   strcat(buffer, "-------------------\n\n");
 
-  for (int m = PROXY_CLIENT; m <= PROXY_SERVER; m++)
+  for (int m = proxy_client; m <= proxy_server; m++)
   {
-    if (m == PROXY_CLIENT)
+    if (m == proxy_client)
     {
       strcat(buffer, "Request\tCached\tSize at Server\t\tSize at Client\t\tCache limit\n");
       strcat(buffer, "-------\t------\t--------------\t\t--------------\t\t-----------\n");
@@ -271,7 +323,7 @@ int Statistics::getClientCacheStats(int type, char *&buffer)
 
     for (int i = 0; i < CHANNEL_STORE_OPCODE_LIMIT; i++)
     {
-      if (m == PROXY_CLIENT)
+      if (m == proxy_client)
       {
         currentStore = proxy_ -> getClientStore() -> getRequestStore(i);
       }
@@ -301,7 +353,7 @@ int Statistics::getClientCacheStats(int type, char *&buffer)
         strcat(buffer, format);
 
         sprintf(format, "%d/%.0f KB\n", currentStore -> cacheSlots,
-                    ((double) control -> getUpperStorageSizeLimit() / 100 *
+                    ((double) control -> getUpperStorageSize() / 100 *
                           currentStore -> cacheThreshold) / 1024);
 
         strcat(buffer, format);
@@ -317,14 +369,14 @@ int Statistics::getClientCacheStats(int type, char *&buffer)
   if (anyStore != NULL)
   {
     sprintf(format, "\ncache: %d bytes (%d KB) available at server.\n",
-                control -> ClientTotalStorageSizeLimit,
-                    control -> ClientTotalStorageSizeLimit / 1024);
+                control -> ClientTotalStorageSize,
+                    control -> ClientTotalStorageSize / 1024);
 
     strcat(buffer, format);
 
     sprintf(format, "       %d bytes (%d KB) available at client.\n\n",
-                control -> ServerTotalStorageSizeLimit,
-                    control -> ServerTotalStorageSizeLimit / 1024);
+                control -> ServerTotalStorageSize,
+                    control -> ServerTotalStorageSize / 1024);
 
     strcat(buffer, format);
 
@@ -398,7 +450,121 @@ int Statistics::getClientProtocolStats(int type, char *&buffer)
   strcat(buffer, "NX Server Side Protocol Statistics\n");
   strcat(buffer, "----------------------------------\n\n");
 
-  strcat(buffer, "Request Total\tCached\tBits In\t\tBits Out\tBits/Request\t\tRatio\n");
+  //
+  // Print render data.
+  //
+
+  strcat(buffer, "Render  Total\tCached\tBits In\t\tBits Out\tBits/Request\t\tRatio\n");
+  strcat(buffer, "------- -----\t------\t-------\t\t--------\t------------\t\t-----\n");
+
+  for (int i = 0; i < STATISTICS_OPCODE_MAX; i++)
+  {
+    if (protocolData -> renderRequestCount_[i])
+    {
+      sprintf(format, "#%d ", i);
+
+      while (strlen(format) < 8)
+      {
+        strcat(format, " ");
+      }
+
+      strcat(buffer, format);
+
+      if (protocolData -> renderRequestCached_[i] > 0)
+      {
+        sprintf(format, "%.0f\t%.0f", protocolData -> renderRequestCount_[i],
+                    protocolData -> renderRequestCached_[i]);
+      }
+      else
+      {
+        sprintf(format, "%.0f\t", protocolData -> renderRequestCount_[i]);
+      }
+
+      strcat(buffer, format);
+
+      sprintf(format, "\t%.0f (%.0f KB)\t%.0f (%.0f KB)\t%.0f/1 -> %.0f/1     \t",
+                  protocolData -> renderRequestBitsIn_[i], protocolData -> renderRequestBitsIn_[i] / 8192,
+                      protocolData -> renderRequestBitsOut_[i], protocolData -> renderRequestBitsOut_[i] / 8192,
+                          protocolData -> renderRequestBitsIn_[i] / protocolData -> renderRequestCount_[i],
+                              protocolData -> renderRequestBitsOut_[i] / protocolData -> renderRequestCount_[i]);
+
+      strcat(buffer, format);
+
+      if (protocolData -> renderRequestBitsOut_[i] > 0)
+      {
+        sprintf(format, "%5.3f:1\n", protocolData -> renderRequestBitsIn_[i] /
+                                         protocolData -> renderRequestBitsOut_[i]);
+
+        strcat(buffer, format);
+      }
+      else
+      {
+        strcat(buffer, "1:1\n");
+      }
+    }
+
+    countRequestIn        += protocolData -> renderRequestCount_[i];
+    countCachedRequestIn  += protocolData -> renderRequestCached_[i];
+
+    countRequestBitsIn  += protocolData -> renderRequestBitsIn_[i];
+    countRequestBitsOut += protocolData -> renderRequestBitsOut_[i];
+
+    countAnyIn   += protocolData -> renderRequestCount_[i];
+    countBitsIn  += protocolData -> renderRequestBitsIn_[i];
+    countBitsOut += protocolData -> renderRequestBitsOut_[i];
+  }
+
+  if (countRequestIn > 0)
+  {
+    if (countCachedRequestIn > 0)
+    {
+      sprintf(format, "\ntotal:  %.0f\t%.0f", countRequestIn, countCachedRequestIn);
+    }
+    else
+    {
+      sprintf(format, "\ntotal:  %.0f\t", countRequestIn);
+    }
+
+    strcat(buffer, format);
+
+    sprintf(format, "\t%.0f (%.0f KB)\t%.0f (%.0f KB)\t%.0f/1 -> %.0f/1     \t",
+                countRequestBitsIn, countRequestBitsIn / 8192, countRequestBitsOut,
+                    countRequestBitsOut / 8192, countRequestBitsIn / countRequestIn,
+                        countRequestBitsOut / countRequestIn);
+
+    strcat(buffer, format);
+
+    if (countRequestBitsOut > 0)
+    {
+      sprintf(format, "%5.3f:1\n", countRequestBitsIn / countRequestBitsOut);
+    }
+    else
+    {
+      sprintf(format, "1:1\n");
+    }
+
+    strcat(buffer, format);
+  }
+  else
+  {
+    strcat(buffer, "N/A\n\n");
+  }
+
+  countRequestIn        = 0;
+  countCachedRequestIn  = 0;
+
+  countRequestBitsIn  = 0;
+  countRequestBitsOut = 0;
+
+  countAnyIn   = 0;
+  countBitsIn  = 0;
+  countBitsOut = 0;
+
+  //
+  // Print other requests' data.
+  //
+
+  strcat(buffer, "\nRequest Total\tCached\tBits In\t\tBits Out\tBits/Request\t\tRatio\n");
   strcat(buffer, "------- -----\t------\t-------\t\t--------\t------------\t\t-----\n");
 
   for (int i = 0; i < STATISTICS_OPCODE_MAX; i++)
@@ -429,7 +595,7 @@ int Statistics::getClientProtocolStats(int type, char *&buffer)
         strcat(format, "R");
       }
 
-      if (i == X_NoOperation && control -> AgentTaintReplies)
+      if (i == X_NoOperation && control -> TaintReplies)
       {
         strcat(format, "T");
       }
@@ -528,22 +694,30 @@ int Statistics::getClientProtocolStats(int type, char *&buffer)
 
   getTimeStats(type, buffer);
 
-  countAnyIn   += protocolData -> syncCount_;
-  countBitsIn  += protocolData -> syncBitsIn_;
-  countBitsOut += protocolData -> syncBitsOut_;
+  countAnyIn   += protocolData -> cupsCount_;
+  countBitsIn  += protocolData -> cupsBitsIn_;
+  countBitsOut += protocolData -> cupsBitsOut_;
 
-  countAnyIn   += protocolData -> keybdCount_;
-  countBitsIn  += protocolData -> keybdBitsIn_;
-  countBitsOut += protocolData -> keybdBitsOut_;
-
-  countAnyIn   += protocolData -> sambaCount_;
-  countBitsIn  += protocolData -> sambaBitsIn_;
-  countBitsOut += protocolData -> sambaBitsOut_;
+  countAnyIn   += protocolData -> smbCount_;
+  countBitsIn  += protocolData -> smbBitsIn_;
+  countBitsOut += protocolData -> smbBitsOut_;
 
   countAnyIn   += protocolData -> mediaCount_;
   countBitsIn  += protocolData -> mediaBitsIn_;
   countBitsOut += protocolData -> mediaBitsOut_;
 
+  countAnyIn   += protocolData -> httpCount_;
+  countBitsIn  += protocolData -> httpBitsIn_;
+  countBitsOut += protocolData -> httpBitsOut_;
+
+  countAnyIn   += protocolData -> fontCount_;
+  countBitsIn  += protocolData -> fontBitsIn_;
+  countBitsOut += protocolData -> fontBitsOut_;
+
+  countAnyIn   += protocolData -> slaveCount_;
+  countBitsIn  += protocolData -> slaveBitsIn_;
+  countBitsOut += protocolData -> slaveBitsOut_;
+
   //
   // Save the overall amount of bytes
   // coming from X clients.
@@ -600,7 +774,7 @@ int Statistics::getClientProtocolStats(int type, char *&buffer)
   getFramingStats(type, buffer);
 
   //
-  // Print sync, samba and audio stats.
+  // Print stats about additional channels.
   //
 
   getServicesStats(type, buffer);
@@ -681,9 +855,9 @@ int Statistics::getClientOverallStats(int type, char *&buffer)
 
   switch (control -> LinkMode)
   {
-    case LINK_TYPE_AUTO:
+    case LINK_TYPE_NONE:
     {
-      strcpy(label, "AUTO");
+      strcpy(label, "NONE");
 
       break;
     }
@@ -805,9 +979,9 @@ int Statistics::getServerCacheStats(int type, char *&buffer)
   strcat(buffer, "\nNX Cache Statistics\n");
   strcat(buffer, "-------------------\n\n");
 
-  for (int m = PROXY_CLIENT; m <= PROXY_SERVER; m++)
+  for (int m = proxy_client; m <= proxy_server; m++)
   {
-    if (m == PROXY_CLIENT)
+    if (m == proxy_client)
     {
       strcat(buffer, "Request\tCached\tSize at Server\t\tSize at Client\t\tCache limit\n");
       strcat(buffer, "-------\t------\t--------------\t\t--------------\t\t-----------\n");
@@ -820,7 +994,7 @@ int Statistics::getServerCacheStats(int type, char *&buffer)
 
     for (int i = 0; i < CHANNEL_STORE_OPCODE_LIMIT; i++)
     {
-      if (m == PROXY_CLIENT)
+      if (m == proxy_client)
       {
         currentStore = proxy_ -> getClientStore() -> getRequestStore(i);
       }
@@ -850,7 +1024,7 @@ int Statistics::getServerCacheStats(int type, char *&buffer)
         strcat(buffer, format);
 
         sprintf(format, "%d/%.0f KB\n", currentStore -> cacheSlots,
-                    ((double) control -> getUpperStorageSizeLimit() / 100 *
+                    ((double) control -> getUpperStorageSize() / 100 *
                           currentStore -> cacheThreshold) / 1024);
 
         strcat(buffer, format);
@@ -866,14 +1040,14 @@ int Statistics::getServerCacheStats(int type, char *&buffer)
   if (anyStore != NULL)
   {
     sprintf(format, "\ncache: %d bytes (%d KB) available at server.\n",
-                control -> ClientTotalStorageSizeLimit,
-                    control -> ClientTotalStorageSizeLimit / 1024);
+                control -> ClientTotalStorageSize,
+                    control -> ClientTotalStorageSize / 1024);
 
     strcat(buffer, format);
 
     sprintf(format, "       %d bytes (%d KB) available at client.\n\n",
-                control -> ServerTotalStorageSizeLimit,
-                    control -> ServerTotalStorageSizeLimit / 1024);
+                control -> ServerTotalStorageSize,
+                    control -> ServerTotalStorageSize / 1024);
 
     strcat(buffer, format);
 
@@ -1181,22 +1355,30 @@ int Statistics::getServerProtocolStats(int type, char *&buffer)
 
   getTimeStats(type, buffer);
 
-  countAnyIn   += protocolData -> syncCount_;
-  countBitsIn  += protocolData -> syncBitsIn_;
-  countBitsOut += protocolData -> syncBitsOut_;
+  countAnyIn   += protocolData -> cupsCount_;
+  countBitsIn  += protocolData -> cupsBitsIn_;
+  countBitsOut += protocolData -> cupsBitsOut_;
 
-  countAnyIn   += protocolData -> keybdCount_;
-  countBitsIn  += protocolData -> keybdBitsIn_;
-  countBitsOut += protocolData -> keybdBitsOut_;
-
-  countAnyIn   += protocolData -> sambaCount_;
-  countBitsIn  += protocolData -> sambaBitsIn_;
-  countBitsOut += protocolData -> sambaBitsOut_;
+  countAnyIn   += protocolData -> smbCount_;
+  countBitsIn  += protocolData -> smbBitsIn_;
+  countBitsOut += protocolData -> smbBitsOut_;
 
   countAnyIn   += protocolData -> mediaCount_;
   countBitsIn  += protocolData -> mediaBitsIn_;
   countBitsOut += protocolData -> mediaBitsOut_;
 
+  countAnyIn   += protocolData -> httpCount_;
+  countBitsIn  += protocolData -> httpBitsIn_;
+  countBitsOut += protocolData -> httpBitsOut_;
+
+  countAnyIn   += protocolData -> fontCount_;
+  countBitsIn  += protocolData -> fontBitsIn_;
+  countBitsOut += protocolData -> fontBitsOut_;
+
+  countAnyIn   += protocolData -> slaveCount_;
+  countBitsIn  += protocolData -> slaveBitsIn_;
+  countBitsOut += protocolData -> slaveBitsOut_;
+
   //
   // Save the overall amount of bytes
   // coming from X clients.
@@ -1253,7 +1435,7 @@ int Statistics::getServerProtocolStats(int type, char *&buffer)
   getFramingStats(type, buffer);
 
   //
-  // Print sync, samba and audio stats.
+  // Print stats about additional channels.
   //
 
   getServicesStats(type, buffer);
@@ -1404,32 +1586,22 @@ int Statistics::getServicesStats(int type, char *&buffer)
 
   char format[FORMAT_LENGTH];
 
-  if (protocolData -> syncBitsOut_ > 0)
+  if (protocolData -> cupsBitsOut_ > 0)
   {
-    sprintf(format, "      %.0f sync messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
-                protocolData -> syncCount_ , protocolData -> syncBitsIn_ / 8,
-                    protocolData -> syncBitsIn_ / 8192, protocolData -> syncBitsOut_ / 8,
-                        protocolData -> syncBitsOut_ / 8192);
+    sprintf(format, "      %.0f CUPS messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
+                protocolData -> cupsCount_ , protocolData -> cupsBitsIn_ / 8,
+                    protocolData -> cupsBitsIn_ / 8192, protocolData -> cupsBitsOut_ / 8,
+                        protocolData -> cupsBitsOut_ / 8192);
 
     strcat(buffer, format);
   }
 
-  if (protocolData -> keybdBitsOut_ > 0)
-  {
-    sprintf(format, "      %.0f embedded keyboard messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
-                protocolData -> keybdCount_ , protocolData -> keybdBitsIn_ / 8,
-                    protocolData -> keybdBitsIn_ / 8192, protocolData -> keybdBitsOut_ / 8,
-                        protocolData -> keybdBitsOut_ / 8192);
-
-    strcat(buffer, format);
-  }
-
-  if (protocolData -> sambaBitsOut_ > 0)
+  if (protocolData -> smbBitsOut_ > 0)
   {
     sprintf(format, "      %.0f SMB messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
-                protocolData -> sambaCount_ , protocolData -> sambaBitsIn_ / 8,
-                    protocolData -> sambaBitsIn_ / 8192, protocolData -> sambaBitsOut_ / 8,
-                        protocolData -> sambaBitsOut_ / 8192);
+                protocolData -> smbCount_ , protocolData -> smbBitsIn_ / 8,
+                    protocolData -> smbBitsIn_ / 8192, protocolData -> smbBitsOut_ / 8,
+                        protocolData -> smbBitsOut_ / 8192);
 
     strcat(buffer, format);
   }
@@ -1444,6 +1616,36 @@ int Statistics::getServicesStats(int type, char *&buffer)
     strcat(buffer, format);
   }
 
+  if (protocolData -> httpBitsOut_ > 0)
+  {
+    sprintf(format, "      %.0f HTTP messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
+                protocolData -> httpCount_ , protocolData -> httpBitsIn_ / 8,
+                    protocolData -> httpBitsIn_ / 8192, protocolData -> httpBitsOut_ / 8,
+                        protocolData -> httpBitsOut_ / 8192);
+
+    strcat(buffer, format);
+  }
+
+  if (protocolData -> fontBitsOut_ > 0)
+  {
+    sprintf(format, "      %.0f font server messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
+                protocolData -> fontCount_ , protocolData -> fontBitsIn_ / 8,
+                    protocolData -> fontBitsIn_ / 8192, protocolData -> fontBitsOut_ / 8,
+                        protocolData -> fontBitsOut_ / 8192);
+
+    strcat(buffer, format);
+  }
+
+  if (protocolData -> slaveBitsOut_ > 0)
+  {
+    sprintf(format, "      %.0f slave messages, %.0f bytes (%.0f KB) in, %.0f bytes (%.0f KB) out.\n\n",
+                protocolData -> slaveCount_ , protocolData -> slaveBitsIn_ / 8,
+                    protocolData -> slaveBitsIn_ / 8192, protocolData -> slaveBitsOut_ / 8,
+                        protocolData -> slaveBitsOut_ / 8192);
+
+    strcat(buffer, format);
+  }
+
   return 1;
 }
 
@@ -1466,8 +1668,9 @@ int Statistics::getFramingStats(int type, char *&buffer)
   // Print info on multiplexing overhead.
   //
 
-  sprintf(format, "%.0f frames in, %.0f frames out.\n\n",
-              transportData -> proxyFrameIn_, transportData -> proxyFrameOut_);
+  sprintf(format, "%.0f frames in, %.0f frames out, %.0f writes out.\n\n",
+              transportData -> proxyFramesIn_, transportData -> proxyFramesOut_,
+                  transportData -> proxyWritesOut_);
 
   strcat(buffer, format);
 
@@ -1555,4 +1758,3 @@ int Statistics::getSplitStats(int type, char *&buffer)
 
   return 1;
 }
-
