@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -15,12 +15,12 @@
 /*                                                                        */
 /**************************************************************************/
 
-#include <X11/Xmd.h>
-
 #include <unistd.h>
 #include <string.h>
 #include <zlib.h>
 
+#include <X11/Xmd.h>
+
 #include "Tight.h"
 
 #include "NXproto.h"
@@ -502,16 +502,17 @@ int Tight::UnpackTight(unsigned char method, T_colormap *new_colormap,
 
   //
   // If the data part of the final image is
-  // bigger than MESSAGE_DATA_LIMIT, cut it
-  // and generate more X_PutImage messages.
+  // bigger than the maximum size allowed
+  // for a normal X request, split it and
+  // generate more X_PutImage messages.
   //
 
   unsigned int dst_size = dst_height * RoundUp4(dst_width * dstBitsPerPixel / 8);
 
-  if (dst_size > MESSAGE_DATA_LIMIT - sz_xPutImageReq)
+  if (dst_size > ((unsigned int) control -> MaximumRequestSize) - 24)
   {
-    unsigned int data_offset = sz_xPutImageReq;
-    unsigned int pieceSize = MESSAGE_DATA_LIMIT - sz_xPutImageReq;
+    unsigned int data_offset = 24;
+    unsigned int pieceSize = control -> MaximumRequestSize - 24;
     unsigned int pieces = 0, currPiece = 0;
 
     if (dst_size % pieceSize == 0)
@@ -962,7 +963,7 @@ int Tight::UnpackTight(T_geometry *geometry, T_colormap *new_colormap,
     }
   }
 
-  const T_colormask *colorMask = GetColorMask(maskMethod);
+  const T_colormask *colorMask = MethodColorMask(maskMethod);
 
   dstBuff = dst_data;
 
