@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -24,7 +24,6 @@
 #include "Transport.h"
 
 #include "Statistics.h"
-#include "Alerts.h"
 
 //
 // Set the verbosity level. You also
@@ -39,6 +38,25 @@
 #undef  INSPECT
 #undef  DUMP
 
+//
+// Define this to lock and unlock the
+// memory-to-memory transport buffers
+// before they are accessed. The code
+// is outdated and doesn't work with
+// the current pthread library.
+//
+
+#undef  THREADS
+
+//
+// Define this to get logging all the
+// operations performed by the parent
+// thread, the one that enqueues and
+// dequeues data.
+//
+
+#define PARENT
+
 //
 // Define this to know when a channel
 // is created or destroyed.
@@ -66,10 +84,12 @@ int InternalTransport::references_;
 Transport::Transport(int fd) : fd_(fd)
 {
   #ifdef TEST
-  *logofs << "Transport: Going to create base transport class "
+  *logofs << "Transport: Going to create base transport "
           << "for FD#" << fd_ << ".\n" << logofs_flush;
   #endif
 
+  type_ = transport_base;
+
   //
   // Set up the write buffer.
   //
@@ -90,7 +110,7 @@ Transport::Transport(int fd) : fd_(fd)
   SetNonBlocking(fd_, 1);
 
   blocked_ = 0;
-  forced_  = 0;
+  drained_ = 0;
   finish_  = 0;
 
   #ifdef REFERENCES
@@ -103,10 +123,12 @@ Transport::Transport(int fd) : fd_(fd)
 Transport::~Transport()
 {
   #ifdef TEST
-  *logofs << "Transport: Going to destroy base socket class "
+  *logofs << "Transport: Going to destroy base class "
           << "for FD#" << fd_ << ".\n" << logofs_flush;
   #endif
 
+  ::close(fd_);
+
   #ifdef REFERENCES
   *logofs << "Transport: Deleted object at " 
           << this << " out of " << --references_ 
@@ -126,16 +148,23 @@ int Transport::read(unsigned char *data, unsigned int size)
   #endif
 
   //
-  // Get any data from socket.
+  // Read the available data from the socket.
   //
 
   int result = ::read(fd_, data, size);
 
+  //
+  // Update the current timestamp as the read
+  // can have scheduled some other process.
+  //
+
+  getNewTimestamp();
+
   if (result < 0)
   {
     if (EGET() == EAGAIN)
     {
-      #if defined(INFO) || defined(TEST)
+      #ifdef TEST
       *logofs << "Transport: WARNING! Read of " << size << " bytes from "
               << "FD#" << fd_ << " would block.\n" << logofs_flush;
       #endif
@@ -203,8 +232,30 @@ int Transport::read(unsigned char *data, unsigned int size)
 int Transport::write(T_write type, const unsigned char *data, const unsigned int size)
 {
   //
-  // If possible, write immediately to the
-  // real socket.
+  // If an immediate write was requested then
+  // flush the enqueued data first.
+  //
+
+  if (w_buffer_.length_ > 0 && blocked_ == 0 &&
+          type == write_immediate)
+  {
+    #ifdef TEST
+    *logofs << "Transport: Writing " << w_buffer_.length_
+            << " bytes of previous data to FD#" << fd_ << ".\n"
+            << logofs_flush;
+    #endif
+
+    int result = Transport::flush();
+
+    if (result < 0)
+    {
+      return -1;
+    }
+  }
+
+  //
+  // If nothing is remained, write immediately
+  // to the socket.
   //
 
   unsigned int written = 0;
@@ -222,7 +273,7 @@ int Transport::write(T_write type, const unsigned char *data, const unsigned int
     #ifdef DUMP
 
     *logofs << "Transport: Going to write " << toWrite
-            << " bytes on FD#" << fd_ << " with checksum ";
+            << " bytes to FD#" << fd_ << " with checksum ";
 
     DumpChecksum(data, size);
 
@@ -244,20 +295,17 @@ int Transport::write(T_write type, const unsigned char *data, const unsigned int
 
       int result = ::write(fd_, data + written, toWrite - written);
 
-      diffTs = diffTimestamp(writeTs, getTimestamp());
+      diffTs = diffTimestamp(writeTs, getNewTimestamp());
 
       control -> addWriteTime(diffTs);
 
-      if (control -> CollectStatistics)
-      {
-        statistics -> addWriteTime(diffTs);
-      }
+      statistics -> addWriteTime(diffTs);
 
       if (result <= 0)
       {
         if (EGET() == EAGAIN)
         {
-          #if defined(INFO) || defined(TEST)
+          #ifdef TEST
           *logofs << "Transport: Write of " << toWrite - written
                   << " bytes on FD#" << fd_ << " would block.\n"
                   << logofs_flush;
@@ -323,22 +371,6 @@ int Transport::write(T_write type, const unsigned char *data, const unsigned int
     return written;
   }
 
-  //
-  // We are going to leave bytes in the write buffer.
-  // This is just the case when user requests delayed
-  // writes. If on the contrary user requested an
-  // immediate write, reason can be that either the
-  // operation would block or that we put a limit on
-  // amount of data to be written. In this case set
-  // the blocked flag to have buffer flushed as soon
-  // as possible.
-  //
-
-  if (type == write_immediate)
-  {
-    blocked_ = 1;
-  }
-
   #ifdef DEBUG
   *logofs << "Transport: Going to append " << size - written
           << " bytes to write buffer for " << "FD#" << fd_
@@ -436,20 +468,17 @@ int Transport::flush()
     int result = ::write(fd_, w_buffer_.data_.begin() + w_buffer_.start_ +
                              written, toWrite - written);
 
-    diffTs = diffTimestamp(writeTs, getTimestamp());
+    diffTs = diffTimestamp(writeTs, getNewTimestamp());
 
     control -> addWriteTime(diffTs);
 
-    if (control -> CollectStatistics)
-    {
-      statistics -> addWriteTime(diffTs);
-    }
+    statistics -> addWriteTime(diffTs);
 
     if (result <= 0)
     {
       if (EGET() == EAGAIN)
       {
-        #if defined(INFO) || defined(TEST)
+        #ifdef TEST
         *logofs << "Transport: Write of " << toWrite - written
                 << " bytes on FD#" << fd_ << " would block.\n"
                 << logofs_flush;
@@ -555,103 +584,108 @@ int Transport::flush()
   return 0;
 }
 
-//
-// Forcibly write data to file descriptor until
-// data left in buffer is below the limit.
-//
-
-int Transport::force(int limit)
+int Transport::drain(int limit, int timeout)
 {
   if (w_buffer_.length_ <= limit)
   {
     return 0;
   }
 
-  forced_ = 1;
+  drained_ = 1;
 
   //
-  // Write only enough data to return under
-  // the limit.
+  // Try to write all the data accumulated.
   //
 
-  int toWrite = w_buffer_.length_ - limit;
+  int toWrite = w_buffer_.length_;
 
-  int written = 0;
+  int written   = 0;
+  int available = 0;
 
-  #if defined(INFO) || defined(TEST)
-  *logofs << "Transport: WARNING! Having to force " << toWrite
+  #ifdef TEST
+  *logofs << "Transport: Draining " << toWrite - limit
           << " bytes on FD#" << fd_ << " with limit set to "
           << limit << ".\n" << logofs_flush;
   #endif
 
-  return 1;
-
-  T_timestamp startTs = getTimestamp();
+  T_timestamp startTs = getNewTimestamp();
 
   T_timestamp selectTs;
   T_timestamp writeTs;
   T_timestamp idleTs;
-  T_timestamp nowTs;
+
+  T_timestamp nowTs = startTs;
 
   int diffTs;
 
   fd_set writeSet;
+  fd_set readSet;
 
   FD_ZERO(&writeSet);
-  FD_SET(fd_, &writeSet);
+  FD_ZERO(&readSet);
 
-  while (written < toWrite)
+  int result;
+
+  while (w_buffer_.length_ - written > limit)
   {
-    writeTs = getTimestamp();
+    writeTs = nowTs;
 
-    int result = ::write(fd_, w_buffer_.data_.begin() + w_buffer_.start_ +
-                             written, toWrite - written);
+    result = ::write(fd_, w_buffer_.data_.begin() + w_buffer_.start_ +
+                         written, toWrite - written);
 
-    nowTs = getTimestamp();
+    nowTs = getNewTimestamp();
 
     diffTs = diffTimestamp(writeTs, nowTs);
 
     control -> addWriteTime(diffTs);
 
-    if (control -> CollectStatistics)
-    {
-      statistics -> addWriteTime(diffTs);
-    }
+    statistics -> addWriteTime(diffTs);
 
     if (result <= 0)
     {
       if (EGET() == EAGAIN)
       {
-        #if defined(INFO) || defined(TEST)
+        #ifdef TEST
         *logofs << "Transport: Write of " << toWrite - written
                 << " bytes on FD#" << fd_ << " would block.\n"
                 << logofs_flush;
         #endif
 
+        //
+        // Check if the result of the last select
+        // told us that there are bytes readable.
+        //
+
+        if (available == 1)
+        {
+          #ifdef WARNING
+          *logofs << "Transport: WARNING! Not draining further "
+                  << "due to data readable on FD#" << fd_
+                  << ".\n" << logofs_flush;
+          #endif
+
+          break;
+        }
+
         //
         // Wait for descriptor to become writable.
         //
 
-        setTimestamp(selectTs, control -> WriteTimeout / 2);
+        FD_SET(fd_, &writeSet);
+        FD_SET(fd_, &readSet);
+
+        setTimestamp(selectTs, timeout / 2);
 
         //
-        // Save a system call using timestamp
-        // got at the end of write.
+        // Save a system call using the timestamp
+        // we got at the end of write.
         //
 
         idleTs = nowTs;
 
-        #if (defined(__hpux) || defined(hpux)) && !defined(PGID_USE_PID)
+        result = select(fd_ + 1, &readSet, &writeSet, NULL, &selectTs);
 
-        select(fd_ + 1, NULL, (int *) &writeSet, NULL, &selectTs);
-
-        #else
-
-        select(fd_ + 1, NULL, &writeSet, NULL, &selectTs);
-
-        #endif
-
-        nowTs = getTimestamp();
+        nowTs = getNewTimestamp();
 
         diffTs = diffTimestamp(idleTs, nowTs);
 
@@ -659,61 +693,48 @@ int Transport::force(int limit)
 
         control -> subReadTime(diffTs);
 
-        if (control -> CollectStatistics)
-        {
-          statistics -> addIdleTime(diffTs);
+        statistics -> addIdleTime(diffTs);
 
-          statistics -> subReadTime(diffTs);
-        }
+        statistics -> subReadTime(diffTs);
 
         diffTs = diffTimestamp(startTs, nowTs);
 
-        if (diffTs >= (control -> WriteTimeout -
-                            (control -> LatencyTimeout * 10)))
+        if (result > 0)
         {
-          int diffSec = (diffTs + control -> LatencyTimeout * 10) / 1000;
-
-          #ifdef PANIC
-          *logofs << "Transport: PANIC! Can't force write to FD#"
-                  << fd_ << " since " << diffSec << " seconds.\n"
-                  << logofs_flush;
-          #endif
-
-          cerr << "Error" << ": Can't write to connection on FD#"
-               << fd_ << " since " << diffSec << " seconds.\n";
-
-          if (control -> ProxyMode == PROXY_CLIENT)
+          if (FD_ISSET(fd_, &readSet))
           {
-            HandleAlert(CLOSE_DEAD_X_CONNECTION_CLIENT_ALERT);
+            #ifdef TEST
+            *logofs << "Transport: The socket for FD#" << fd_
+                    << " became readable.\n" << logofs_flush;
+            #endif
+
+            available = 1;
+
+            if (result == 1)
+            {
+              #ifdef WARNING
+              *logofs << "Transport: WARNING! Not draining further "
+                      << "due to data readable on FD#" << fd_
+                      << ".\n" << logofs_flush;
+              #endif
+
+              break;
+            }
           }
-          else
-          {
-            HandleAlert(CLOSE_DEAD_X_CONNECTION_SERVER_ALERT);
-          }
-
-          #ifdef PANIC
-          *logofs << "Transport: PANIC! Shutting down dead connection "
-                  << "on FD#" << fd_ << ".\n" << logofs_flush;
-          #endif
-
-          cerr << "Error" << ": Shutting down dead connection "
-               << "on FD#" << fd_ << ".\n";
-
-          shutdown(fd_, SHUT_RDWR);
-
-          break;
         }
-        #ifdef WARNING
-        else if (diffTs >= (control -> PingTimeout -
-                                 (control -> LatencyTimeout * 10)))
+        else if (result <= 0)
         {
-          int diffSec = (diffTs + control -> LatencyTimeout * 10) / 1000;
+          if (diffTs >= timeout)
+          {
+            #ifdef WARNING
+            *logofs << "Transport: WARNING! Not draining further "
+                    << "due to the timeout on FD#" << fd_
+                    << ".\n" << logofs_flush;
+            #endif
 
-          *logofs << "Transport: WARNING: Couldn't force write to FD#"
-                  << fd_ << " for " << diffSec << " seconds.\n"
-                  << logofs_flush;
+            break;
+          }
         }
-        #endif
       }
       else if (EGET() == EINTR)
       {
@@ -747,6 +768,17 @@ int Transport::force(int limit)
 
       written += result;
     }
+
+    if (available == 1)
+    {
+      #ifdef WARNING
+      *logofs << "Transport: WARNING! Not draining further "
+              << "due to data readable on FD#" << fd_
+              << ".\n" << logofs_flush;
+      #endif
+
+      break;
+    }
   }
 
   if (written > 0)
@@ -788,8 +820,8 @@ int Transport::force(int limit)
   #ifdef TEST
   else
   {
-    *logofs << "Transport: WARNING! No data written on FD#" << fd_
-            << " with " << toWrite  << " bytes to force and limit "
+    *logofs << "Transport: WARNING! No data written to FD#" << fd_
+            << " with " << toWrite  << " bytes to drain and limit "
             << "set to " << limit << ".\n" << logofs_flush;
   }
   #endif
@@ -806,12 +838,12 @@ int Transport::force(int limit)
           << logofs_flush;
   #endif
 
-  return 1;
+  return (w_buffer_.length_ <= limit);
 }
 
 int Transport::wait(int timeout) const
 {
-  T_timestamp startTs = getTimestamp();
+  T_timestamp startTs = getNewTimestamp();
 
   T_timestamp idleTs;
   T_timestamp selectTs;
@@ -835,16 +867,15 @@ int Transport::wait(int timeout) const
     diffTs = diffTimestamp(startTs, nowTs);
 
     if (available != 0 || timeout == 0 ||
-            (diffTs + timeout / 10) >= timeout)
+            (diffTs + (timeout / 10)) >= timeout)
     {
       #ifdef TEST
       *logofs << "Transport: There are " << available
               << " bytes on FD#" << fd_ << " after "
-              << diffTimestamp(startTs, getTimestamp())
-              << " Ms.\n" << logofs_flush;
+              << diffTs << " Ms.\n" << logofs_flush;
       #endif
 
-      return (int) available;
+      return available;
     }
     else if (available == 0 && result > 0)
     {
@@ -864,23 +895,15 @@ int Transport::wait(int timeout) const
     selectTs.tv_sec  = 0;
     selectTs.tv_usec = timeout * 1000;
 
-    idleTs = getTimestamp();
+    idleTs = nowTs;
 
     //
     // Wait for descriptor to become readable.
     //
 
-    #if (defined(__hpux) || defined(hpux)) && !defined(PGID_USE_PID)
-
-    result = select(fd_ + 1, (int *) &readSet, NULL, NULL, &selectTs);
-
-    #else
-
     result = select(fd_ + 1, &readSet, NULL, NULL, &selectTs);
 
-    #endif
-
-    nowTs = getTimestamp();
+    nowTs = getNewTimestamp();
 
     diffTs = diffTimestamp(idleTs, nowTs);
 
@@ -888,12 +911,9 @@ int Transport::wait(int timeout) const
 
     control -> subReadTime(diffTs);
 
-    if (control -> CollectStatistics)
-    {
-      statistics -> addIdleTime(diffTs);
+    statistics -> addIdleTime(diffTs);
 
-      statistics -> subReadTime(diffTs);
-    }
+    statistics -> subReadTime(diffTs);
 
     if (result < 0)
     {
@@ -920,13 +940,13 @@ int Transport::wait(int timeout) const
     else if (result == 0)
     {
       *logofs << "Transport: No data available on FD#" << fd_
-              << " after " << diffTimestamp(startTs, getTimestamp())
+              << " after " << diffTimestamp(startTs, nowTs)
               << " Ms.\n" << logofs_flush;
     }
     else
     {
       *logofs << "Transport: Data became available on FD#" << fd_
-              << " after " << diffTimestamp(startTs, getTimestamp())
+              << " after " << diffTimestamp(startTs, nowTs)
               << " Ms.\n" << logofs_flush;
     }
     #endif
@@ -950,7 +970,7 @@ void Transport::setSize(unsigned int initialSize, unsigned int thresholdSize,
 int Transport::fullReset()
 {
   blocked_ = 0;
-  forced_  = 0;
+  drained_ = 0;
   finish_  = 0;
 
   return fullReset(w_buffer_);
@@ -1027,26 +1047,7 @@ int Transport::resize(T_buffer &buffer, const int &size)
             << " bytes.\n" << logofs_flush;
     #endif
 
-    try
-    {
-      buffer.data_.resize(newSize);
-    }
-    catch (exception e)
-    {
-      #ifdef PANIC
-      *logofs << "Transport: PANIC! Buffer resize for FD#" << fd_
-              << " from " << buffer.data_.size() << " to " << newSize
-              << " bytes failed.\n" << logofs_flush;
-      #endif
-
-      cerr << "Error" << ": Buffer resize for FD#" << fd_
-           << " from " << buffer.data_.size() << " to " << newSize
-           << " bytes failed.\n";
-
-      finish();
-
-      return -1;
-    }
+    buffer.data_.resize(newSize);
 
     #ifdef TEST
     if (newSize >= maximumSize_)
@@ -1075,6 +1076,13 @@ int Transport::fullReset(T_buffer &buffer)
   // of the initial size.
   //
 
+  #ifdef TEST
+  *logofs << "Transport: Resetting buffer for " << "FD#"
+          << fd_ << " with size " << buffer.data_.size()
+          << " and capacity " << buffer.data_.capacity()
+          << ".\n" << logofs_flush;
+  #endif
+
   buffer.start_  = 0;
   buffer.length_ = 0;
 
@@ -1084,26 +1092,9 @@ int Transport::fullReset(T_buffer &buffer)
     return 0;
   }
 
-  try
-  {
-    buffer.data_.clear();
+  buffer.data_.clear();
 
-    buffer.data_.resize(initialSize_);
-  }
-  catch (exception e)
-  {
-    #ifdef PANIC
-    *logofs << "Transport: PANIC! Buffer shrink failed for "
-            << "FD#" << fd_ << ".\n" << logofs_flush;
-    #endif
-
-    cerr << "Error" << ": Buffer shrink failed for "
-         << "FD#" << fd_ << ".\n";
-
-    finish();
-
-    return -1;
-  }
+  buffer.data_.resize(initialSize_);
 
   #ifdef TEST
   *logofs << "Transport: Data buffer for " << "FD#"
@@ -1117,6 +1108,13 @@ int Transport::fullReset(T_buffer &buffer)
 
 ProxyTransport::ProxyTransport(int fd) : Transport(fd)
 {
+  #ifdef TEST
+  *logofs << "ProxyTransport: Going to create proxy transport "
+          << "for FD#" << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  type_ = transport_proxy;
+
   //
   // Set up the read buffer.
   //
@@ -1188,8 +1186,13 @@ ProxyTransport::ProxyTransport(int fd) : Transport(fd)
 
 ProxyTransport::~ProxyTransport()
 {
+  #ifdef TEST
+  *logofs << "ProxyTransport: Going to destroy derived class "
+          << "for FD#" << fd_ << ".\n" << logofs_flush;
+  #endif
+
   //
-  // Free ZLIB dynamic structures.
+  // Deallocate the ZLIB stream state.
   //
 
   inflateEnd(&r_stream_);
@@ -1228,10 +1231,7 @@ int ProxyTransport::read(unsigned char *data, unsigned int size)
 
     control -> addBytesIn(result);
 
-    if (control -> CollectStatistics)
-    {
-      statistics -> addBytesIn(result);
-    }
+    statistics -> addBytesIn(result);
 
     return result;
   }
@@ -1290,10 +1290,7 @@ int ProxyTransport::read(unsigned char *data, unsigned int size)
 
   control -> addBytesIn(result);
 
-  if (control -> CollectStatistics)
-  {
-    statistics -> addBytesIn(result);
-  }
+  statistics -> addBytesIn(result);
 
   //
   // Uncompress data into the read buffer.
@@ -1497,7 +1494,7 @@ int ProxyTransport::read(unsigned char *data, unsigned int size)
     }
 
     #ifdef TEST
-    *logofs << "ProxyTransport: Need to add " <<  newAvailOut
+    *logofs << "ProxyTransport: Need to add " << newAvailOut
             << " bytes to decompress buffer.\n"
             << logofs_flush;
     #endif
@@ -1512,10 +1509,7 @@ int ProxyTransport::read(unsigned char *data, unsigned int size)
           << " bytes.\n" << logofs_flush;
   #endif
 
-  if (control -> CollectStatistics)
-  {
-    statistics -> addDecompressedBytes(diffTotalIn, diffTotalOut);
-  }
+  statistics -> addDecompressedBytes(diffTotalIn, diffTotalOut);
 
   //
   // Copy the uncompressed data to the provided buffer.
@@ -1589,12 +1583,9 @@ int ProxyTransport::write(T_write type, const unsigned char *data, const unsigne
 
     control -> addBytesOut(result);
 
-    if (control -> CollectStatistics)
-    {
-      statistics -> addBytesOut(result);
-    }
+    statistics -> addBytesOut(result);
 
-    control -> setLastWriteInfo(result);
+    control -> updateBitrate(result);
 
     return result;
   }
@@ -1806,7 +1797,7 @@ int ProxyTransport::write(T_write type, const unsigned char *data, const unsigne
     }
 
     #ifdef TEST
-    *logofs << "ProxyTransport: Need to add " <<  newAvailOut
+    *logofs << "ProxyTransport: Need to add " << newAvailOut
             << " bytes to compress buffer.\n"
             << logofs_flush;
     #endif
@@ -1833,32 +1824,7 @@ int ProxyTransport::write(T_write type, const unsigned char *data, const unsigne
   #endif
 
   //
-  // We added bytes to write buffer. Data is not
-  // written to network yet, but it's convenient
-  // to update the counters at this stage to get
-  // the current bitrate earlier.
-  //
-
-  control -> addBytesOut(diffTotalOut);
-
-  if (control -> CollectStatistics)
-  {
-    statistics -> addCompressedBytes(diffTotalIn, diffTotalOut);
-
-    statistics -> addBytesOut(diffTotalOut);
-  }
-
-  //
-  // Update the current bitrate.
-  //
-
-  if (diffTotalOut > 0)
-  {
-    control -> setLastWriteInfo(diffTotalOut);
-  }
-
-  //
-  // Find out what to do with the
+  // Find out what we have to do with the
   // produced data.
   //
 
@@ -1911,6 +1877,21 @@ int ProxyTransport::write(T_write type, const unsigned char *data, const unsigne
     flush_ += diffTotalIn;
   }
 
+  //
+  // We either wrote the data or added it to the
+  // write buffer. It's convenient to update the
+  // counters at this stage to get the current
+  // bitrate earlier.
+  //
+
+  control -> addBytesOut(diffTotalOut);
+
+  statistics -> addCompressedBytes(diffTotalIn, diffTotalOut);
+
+  statistics -> addBytesOut(diffTotalOut);
+
+  control -> updateBitrate(diffTotalOut);
+
   #ifdef TEST
   *logofs << "ProxyTransport: Write buffer for proxy FD#" << fd_
           << " has data for " << w_buffer_.length_ << " bytes.\n"
@@ -1961,12 +1942,6 @@ int ProxyTransport::flush()
           << " is " << flush_ << " bytes.\n" << logofs_flush;
   #endif
 
-  //
-  // Reset the blocked flag.
-  //
-
-  blocked_ = 0;
-
   //
   // Flush ZLIB stream into the write buffer.
   //
@@ -2167,7 +2142,7 @@ int ProxyTransport::flush()
     }
 
     #ifdef TEST
-    *logofs << "ProxyTransport: Need to add " <<  newAvailOut
+    *logofs << "ProxyTransport: Need to add " << newAvailOut
             << " bytes to compress buffer in flush.\n"
             << logofs_flush;
     #endif
@@ -2182,24 +2157,6 @@ int ProxyTransport::flush()
           << " bytes.\n" << logofs_flush;
   #endif
 
-  control -> addBytesOut(diffTotalOut);
-
-  if (control -> CollectStatistics)
-  {
-    statistics -> addCompressedBytes(diffTotalIn, diffTotalOut);
-
-    statistics -> addBytesOut(diffTotalOut);
-  }
-
-  //
-  // Update the current bitrate.
-  //
-
-  if (diffTotalOut > 0)
-  {
-    control -> setLastWriteInfo(diffTotalOut);
-  }
-
   //
   // Time to move data from the write
   // buffer to the real link. 
@@ -2231,13 +2188,25 @@ int ProxyTransport::flush()
     return -1;
   }
 
+  //
+  // Update all the ounters.
+  //
+
+  control -> addBytesOut(diffTotalOut);
+
+  statistics -> addCompressedBytes(diffTotalIn, diffTotalOut);
+
+  statistics -> addBytesOut(diffTotalOut);
+
+  control -> updateBitrate(diffTotalOut);
+
   return result;
 }
 
 int ProxyTransport::fullReset()
 {
   blocked_ = 0;
-  forced_  = 0;
+  drained_ = 0;
   finish_  = 0;
   flush_   = 0;
 
@@ -2267,3 +2236,743 @@ int ProxyTransport::fullReset()
 
   return (result == 4);
 }
+
+AgentTransport::AgentTransport(int fd) : Transport(fd)
+{
+  #ifdef TEST
+  *logofs << "AgentTransport: Going to create agent transport "
+          << "for FD#" << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  type_ = transport_agent;
+
+  //
+  // Set up the read buffer.
+  //
+
+  r_buffer_.length_ = 0;
+  r_buffer_.start_  = 0;
+
+  r_buffer_.data_.resize(initialSize_);
+
+  //
+  // Set up the mutexes.
+  //
+
+  #ifdef THREADS
+
+  pthread_mutexattr_t m_attributes;
+
+  pthread_mutexattr_init(&m_attributes);
+
+  //
+  // Interfaces in pthread to handle mutex
+  // type do not work in current version. 
+  //
+
+  m_attributes.__mutexkind = PTHREAD_MUTEX_ERRORCHECK_NP;
+
+  if (pthread_mutex_init(&m_read_, &m_attributes) != 0)
+  {
+    #ifdef TEST
+    *logofs << "AgentTransport: Child: Creation of read mutex failed. "
+            << "Error is " << EGET() << " '" << ESTR()
+            << "'.\n" << logofs_flush;
+    #endif
+  }
+
+  if (pthread_mutex_init(&m_write_, &m_attributes) != 0)
+  {
+    #ifdef TEST
+    *logofs << "AgentTransport: Child: Creation of write mutex failed. "
+            << "Error is " << EGET() << " '" << ESTR()
+            << "'.\n" << logofs_flush;
+    #endif
+  }
+
+  #endif
+
+  #ifdef REFERENCES
+  *logofs << "AgentTransport: Child: Created new object at " 
+          << this << " out of " << ++references_ 
+          << " allocated references.\n" << logofs_flush;
+  #endif
+}
+
+AgentTransport::~AgentTransport()
+{
+  #ifdef TEST
+  *logofs << "AgentTransport: Going to destroy derived class "
+          << "for FD#" << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  //
+  // Unlock and free all mutexes.
+  //
+
+  #ifdef THREADS
+
+  pthread_mutex_unlock(&m_read_);
+  pthread_mutex_unlock(&m_write_);
+
+  pthread_mutex_destroy(&m_read_);
+  pthread_mutex_destroy(&m_write_);
+
+  #endif
+
+  #ifdef REFERENCES
+  *logofs << "AgentTransport: Child: Deleted object at " 
+          << this << " out of " << --references_ 
+          << " allocated references.\n" << logofs_flush;
+  #endif
+}
+
+//
+// Read data enqueued by the other thread.
+//
+
+int AgentTransport::read(unsigned char *data, unsigned int size)
+{
+  #ifdef THREADS
+
+  lockRead();
+
+  #endif
+
+  #ifdef DEBUG
+  *logofs << "AgentTransport: Child: Going to read " << size
+          << " bytes from " << "FD#" << fd_ << ".\n"
+          << logofs_flush;
+  #endif
+
+  int copied = 0;
+
+  if (r_buffer_.length_ > 0)
+  {
+    copied = (r_buffer_.length_ > ((int) size) ?
+                  ((int) size) : r_buffer_.length_);
+
+    memcpy(data, r_buffer_.data_.begin() + r_buffer_.start_, copied);
+
+    //
+    // Update buffer status.
+    //
+
+    #ifdef TEST
+    *logofs << "AgentTransport: Child: Going to immediately return "
+            << copied << " bytes from FD#" << fd_ << ".\n"
+            << logofs_flush;
+    #endif
+
+    #ifdef DUMP
+
+    *logofs << "AgentTransport: Child: Dumping content of read data.\n"
+            << logofs_flush;
+
+    DumpData(data, copied);
+
+    #endif
+
+    r_buffer_.length_ -= copied;
+
+    if (r_buffer_.length_ == 0)
+    {
+      r_buffer_.start_ = 0;
+    }
+    else
+    {
+      r_buffer_.start_ += copied;
+
+      #ifdef TEST
+      *logofs << "AgentTransport: Child: There are still "
+              << r_buffer_.length_  << " bytes in read buffer for "
+              << "FD#" << fd_ << ".\n" << logofs_flush;
+      #endif
+    }
+  }
+  else
+  {
+    #ifdef DEBUG
+    *logofs << "AgentTransport: Child: No data can be gotten "
+            << "from read buffer for FD#" << fd_ << ".\n"
+            << logofs_flush;
+    #endif
+
+    ESET(EAGAIN);
+  }
+
+  #ifdef THREADS
+
+  unlockRead();
+
+  #endif
+
+  //
+  // Update the timestamp because the
+  // agent may have run for a relevant
+  // amount of time.
+  //
+
+  getNewTimestamp();
+
+  return copied;
+}
+
+//
+// Write data to buffer so that the other
+// thread can get it.
+//
+
+int AgentTransport::write(T_write type, const unsigned char *data, const unsigned int size)
+{
+  #ifdef THREADS
+
+  lockWrite();
+
+  #endif
+
+  //
+  // Just append data to socket's write buffer.
+  // Note that we don't care if buffer exceeds
+  // the size limits set for this type of
+  // transport.
+  //
+
+  #ifdef TEST
+  *logofs << "AgentTransport: Child: Going to append " << size
+          << " bytes to write buffer for " << "FD#" << fd_
+          << ".\n" << logofs_flush;
+  #endif
+
+  int copied = 0;
+
+  if (resize(w_buffer_, size) < 0)
+  {
+    finish();
+
+    ESET(EPIPE);
+
+    copied = -1;
+  }
+  else
+  {
+    memmove(w_buffer_.data_.begin() + w_buffer_.start_ + w_buffer_.length_, data, size);
+
+    w_buffer_.length_ += size;
+
+    #ifdef DUMP
+
+    *logofs << "AgentTransport: Child: Dumping content of written data.\n"
+            << logofs_flush;
+
+    DumpData(data, size);
+
+    #endif
+
+    #ifdef TEST
+    *logofs << "AgentTransport: Child: Write buffer for FD#" << fd_
+            << " has data for " << w_buffer_.length_ << " bytes.\n"
+            << logofs_flush;
+
+    *logofs << "AgentTransport: Child: Start is " << w_buffer_.start_ 
+            << " length is " << w_buffer_.length_ << " size is "
+            << w_buffer_.data_.size() << " capacity is "
+            << w_buffer_.data_.capacity() << ".\n"
+            << logofs_flush;
+    #endif
+
+    copied = size;
+  }
+
+  //
+  // Let child access again the read buffer.
+  //
+
+  #ifdef THREADS
+
+  unlockWrite();
+
+  #endif
+
+  return copied;
+}
+
+int AgentTransport::flush()
+{
+  //
+  // In case of memory-to-memory transport
+  // this function should never be called.
+  //
+
+  #ifdef PANIC
+  *logofs << "AgentTransport: Child: PANIC! Called flush() for "
+          << "memory to memory transport on " << "FD#"
+          << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  cerr << "Error" << ": Called flush() for "
+       << "memory to memory transport on " << "FD#"
+       << fd_ << ".\n";
+
+  HandleAbort();
+}
+
+int AgentTransport::drain(int limit, int timeout)
+{
+  //
+  // This function should be never called in
+  // the case of memory-to-memory transport.
+  // Data is enqueued for the agent to read
+  // but the agent could require multiple
+  // loops to read it all.
+  //
+
+  #ifdef TEST
+  *logofs << "AgentTransport: Child: WARNING! Called drain() for "
+          << "memory to memory transport on " << "FD#"
+          << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  return 0;
+}
+
+int AgentTransport::writable() const
+{
+  #if defined(PARENT) && (defined(TEST) || defined(INFO))
+  *logofs << "AgentTransport: Child: WARNING! Called for "
+          << "FD#" << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  int writable = control -> TransportXBufferThreshold -
+                     w_buffer_.length_;
+
+  if (writable < 0)
+  {
+    writable = 0;
+  }
+
+  #if defined(PARENT) && (defined(TEST) || defined(INFO))
+  if (writable <= 0)
+  {
+    *logofs << "AgentTransport: Child: WARNING! No data is "
+            << "writable on write buffer for FD#" << fd_
+            << ".\n" << logofs_flush;
+  }
+  #endif
+
+  return writable;
+}
+
+int AgentTransport::fullReset()
+{
+  #ifdef THREADS
+
+  lockRead();
+  lockWrite();
+
+  #endif
+
+  #ifdef TEST
+  *logofs << "AgentTransport: Child: Resetting transport "
+          << "for FD#" << fd_ << ".\n" << logofs_flush;
+  #endif
+
+  blocked_ = 0;
+  drained_ = 0;
+  finish_  = 0;
+
+  int result = 0;
+
+  if (Transport::fullReset(r_buffer_) >= 0)
+  {
+    result++;
+  }
+
+  if (Transport::fullReset(w_buffer_) >= 0)
+  {
+    result++;
+  }
+
+  return (result == 2);
+}
+
+int AgentTransport::enqueue(const char *data, const int size)
+{
+  #ifdef THREADS
+
+  lockRead();
+
+  #endif
+
+  if (finish_ == 1)
+  {
+    #if defined(PARENT) && defined(TEST)
+    *logofs << "AgentTransport: Parent: Returning EPIPE in "
+            << "write for finishing FD#" << fd_ << ".\n"
+           << logofs_flush;
+    #endif
+
+    ESET(EPIPE);
+
+    return -1;
+  }
+
+  //
+  // Don't allow the agent to write more data
+  // than the threshold.
+  //
+
+  int toPut = control -> TransportXBufferThreshold -
+                  r_buffer_.length_;
+
+  if (toPut <= 0)
+  {
+    #if defined(PARENT) && defined(TEST)
+    *logofs << "AgentTransport: Parent: WARNING! Read buffer for FD#"
+            << fd_ << " has already reached the limit of "
+            << control -> TransportXBufferThreshold
+            << " bytes.\n" << logofs_flush;
+    #endif
+
+    ESET(EAGAIN);
+
+    #ifdef THREADS
+
+    unlockRead();
+
+    #endif
+
+    return -1;
+  }
+
+  toPut = (toPut > size ? size : toPut);
+
+  #if defined(PARENT) && defined(TEST)
+  *logofs << "AgentTransport: Parent: Going to put " << toPut
+          << " bytes into read buffer for FD#" << fd_
+          << ". Buffer length is " << r_buffer_.length_
+          << ".\n" << logofs_flush;
+  #endif
+
+  if (resize(r_buffer_, toPut) < 0)
+  {
+    finish();
+
+    #ifdef THREADS
+
+    unlockRead();
+
+    #endif
+
+    return -1;
+  }
+
+  memcpy(r_buffer_.data_.begin() + r_buffer_.start_ + r_buffer_.length_, data, toPut);
+
+  r_buffer_.length_ += toPut;
+
+  #if defined(DUMP) && defined(PARENT)
+
+  *logofs << "AgentTransport: Parent: Dumping content of enqueued data.\n"
+          << logofs_flush;
+
+  DumpData(data, toPut);
+
+  #endif
+
+  #if defined(PARENT) && defined(TEST)
+  *logofs << "AgentTransport: Parent: Read buffer for FD#" << fd_
+          << " has now data for " << r_buffer_.length_
+          << " bytes.\n" << logofs_flush;
+
+  *logofs << "AgentTransport: Parent: Start is " << r_buffer_.start_ 
+          << " length is " << r_buffer_.length_ << " size is "
+          << r_buffer_.data_.size() << " capacity is "
+          << r_buffer_.data_.capacity() << ".\n"
+          << logofs_flush;
+  #endif
+
+  #ifdef THREADS
+
+  unlockRead();
+
+  #endif
+
+  return toPut;
+}
+
+int AgentTransport::dequeue(char *data, int size)
+{
+  #ifdef THREADS
+
+  lockWrite();
+
+  #endif
+
+  if (w_buffer_.length_ == 0)
+  {
+    if (finish_ == 1)
+    {
+      #if defined(PARENT) && defined(TEST)
+      *logofs << "AgentTransport: Parent: Returning 0 in read "
+              << "for finishing FD#" << fd_ << ".\n"
+              << logofs_flush;
+      #endif
+
+      return 0;
+    }
+
+    #if defined(PARENT) && defined(TEST)
+    *logofs << "AgentTransport: Parent: No data can be read "
+            << "from write buffer for FD#" << fd_ << ".\n"
+            << logofs_flush;
+    #endif
+
+    ESET(EAGAIN);
+
+    #ifdef THREADS
+
+    unlockWrite();
+
+    #endif
+
+    return -1;
+  }
+
+  //
+  // Return as many bytes as possible.
+  //
+
+  int toGet = ((int) size > w_buffer_.length_ ? w_buffer_.length_ : size);
+
+  #if defined(PARENT) && defined(TEST)
+  *logofs << "AgentTransport: Parent: Going to get " << toGet
+          << " bytes from write buffer for FD#" << fd_ << ".\n"
+          << logofs_flush;
+  #endif
+
+  memcpy(data, w_buffer_.data_.begin() + w_buffer_.start_, toGet);
+
+  w_buffer_.start_  += toGet;
+  w_buffer_.length_ -= toGet;
+
+  #if defined(DUMP) && defined(PARENT)
+
+  *logofs << "AgentTransport: Parent: Dumping content of dequeued data.\n"
+          << logofs_flush;
+
+  DumpData(data, toGet);
+
+  #endif
+
+  #if defined(PARENT) && defined(TEST)
+  *logofs << "AgentTransport: Parent: Write buffer for FD#" << fd_
+          << " has now data for " << length() << " bytes.\n"
+          << logofs_flush;
+
+  *logofs << "AgentTransport: Parent: Start is " << w_buffer_.start_ 
+          << " length is " << w_buffer_.length_ << " size is "
+          << w_buffer_.data_.size() << " capacity is "
+          << w_buffer_.data_.capacity() << ".\n"
+          << logofs_flush;
+  #endif
+
+  #ifdef THREADS
+
+  unlockWrite();
+
+  #endif
+
+  return toGet;
+}
+
+int AgentTransport::queuable()
+{
+  #ifdef THREADS
+
+  lockRead();
+
+  #endif
+
+  int queuable = control -> TransportXBufferThreshold -
+                     r_buffer_.length_;
+
+  if (queuable < 0)
+  {
+    queuable = 0;
+  }
+
+  #if defined(PARENT) && defined(TEST)
+  if (queuable <= 0)
+  {
+    *logofs << "AgentTransport: Parent: No data is queuable "
+            << "on read buffer for FD#" << fd_ << ".\n"
+            << logofs_flush;
+  }
+  #endif
+
+  #ifdef THREADS
+
+  unlockRead();
+
+  #endif
+
+  return queuable;
+}
+
+int AgentTransport::dequeuable()
+{
+  if (finish_ == 1)
+  {
+    #if defined(PARENT) && defined(TEST)
+    *logofs << "AgentTransport: Parent: Returning EPIPE in "
+            << "readable for finishing FD#" << fd_
+            << ".\n" << logofs_flush;
+    #endif
+
+    ESET(EPIPE);
+
+    return -1;
+  }
+
+  #if defined(PARENT) && defined(TEST)
+  *logofs << "AgentTransport: Parent: Returning "
+          << w_buffer_.length_ << " as data readable "
+          << "from read buffer for FD#" << fd_ << ".\n"
+          << logofs_flush;
+  #endif
+
+  return w_buffer_.length_;
+}
+
+#ifdef THREADS
+
+int AgentTransport::lockRead()
+{
+  for (;;)
+  {
+    int result = pthread_mutex_lock(&m_read_);
+
+    if (result == 0)
+    {
+      #ifdef DEBUG
+      *logofs << "AgentTransport: Read mutex locked by thread id "
+              << pthread_self() << ".\n" << logofs_flush;
+      #endif
+
+      return 0;
+    }
+    else if (EGET() == EINTR)
+    {
+      continue;
+    }
+    else
+    {
+      #ifdef WARNING
+      *logofs << "AgentTransport: WARNING! Locking of read mutex by thread id "
+              << pthread_self() << " returned " << result << ". Error is '"
+              << ESTR() << "'.\n" << logofs_flush;
+      #endif
+
+      return result;
+    }
+  }
+}
+
+int AgentTransport::lockWrite()
+{
+  for (;;)
+  {
+    int result = pthread_mutex_lock(&m_write_);
+
+    if (result == 0)
+    {
+      #ifdef DEBUG
+      *logofs << "AgentTransport: Write mutex locked by thread id "
+              << pthread_self() << ".\n" << logofs_flush;
+      #endif
+
+      return 0;
+    }
+    else if (EGET() == EINTR)
+    {
+      continue;
+    }
+    else
+    {
+      #ifdef WARNING
+      *logofs << "AgentTransport: WARNING! Locking of write mutex by thread id "
+              << pthread_self() << " returned " << result << ". Error is '"
+              << ESTR() << "'.\n" << logofs_flush;
+      #endif
+
+      return result;
+    }
+  }
+}
+
+int AgentTransport::unlockRead()
+{
+  for (;;)
+  {
+    int result = pthread_mutex_unlock(&m_read_);
+
+    if (result == 0)
+    {
+      #ifdef DEBUG
+      *logofs << "AgentTransport: Read mutex unlocked by thread id "
+              << pthread_self() << ".\n" << logofs_flush;
+      #endif
+
+      return 0;
+    }
+    else if (EGET() == EINTR)
+    {
+      continue;
+    }
+    else
+    {
+      #ifdef WARNING
+      *logofs << "AgentTransport: WARNING! Unlocking of read mutex by thread id "
+              << pthread_self() << " returned " << result << ". Error is '"
+              << ESTR() << "'.\n" << logofs_flush;
+      #endif
+
+      return result;
+    }
+  }
+}
+
+int AgentTransport::unlockWrite()
+{
+  for (;;)
+  {
+    int result = pthread_mutex_unlock(&m_write_);
+
+    if (result == 0)
+    {
+      #ifdef DEBUG
+      *logofs << "AgentTransport: Write mutex unlocked by thread id "
+              << pthread_self() << ".\n" << logofs_flush;
+      #endif
+
+      return 0;
+    }
+    else if (EGET() == EINTR)
+    {
+      continue;
+    }
+    else
+    {
+      #ifdef WARNING
+      *logofs << "AgentTransport: WARNING! Unlocking of write mutex by thread id "
+              << pthread_self() << " returned " << result << ". Error is '"
+              << ESTR() << "'.\n" << logofs_flush;
+      #endif
+
+      return result;
+    }
+  }
+}
+
+#endif
