@@ -1,6 +1,6 @@
 /**************************************************************************/
 /*                                                                        */
-/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
+/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
 /*                                                                        */
 /* NXPROXY, NX protocol compression and NX extensions to this software    */
 /* are copyright of NoMachine. Redistribution and use of the present      */
@@ -23,25 +23,11 @@ XidCache::XidCache()
 {
   for (int i = 0; i < 256; i++)
   {
-    if (control -> isProtoStep2() == 1)
-    {
-      base_[i] = new IntCache(8);
-    }
-    else
-    {
-      base_[i] = new IntCache(32);
-    }
+    base_[i] = new IntCache(8);
   }
 
   slot_ = 0;
   last_ = 0;
-
-  //
-  // Only used in protocol
-  // version less than 1.2.
-  //
-
-  mask_ = 0;
 }
 
 XidCache::~XidCache()
